import json
import logging
from io import StringIO

from letta.log import JSONFormatter, LogContextFilter
from letta.log_context import clear_log_context, get_log_context, remove_log_context, set_log_context, update_log_context


class TestLogContext:
    def test_set_log_context(self):
        clear_log_context()
        set_log_context("agent_id", "agent-123")
        assert get_log_context("agent_id") == "agent-123"
        clear_log_context()

    def test_update_log_context(self):
        clear_log_context()
        update_log_context(agent_id="agent-123", actor_id="user-456")
        context = get_log_context()
        assert context["agent_id"] == "agent-123"
        assert context["actor_id"] == "user-456"
        clear_log_context()

    def test_remove_log_context(self):
        clear_log_context()
        update_log_context(agent_id="agent-123", actor_id="user-456")
        remove_log_context("agent_id")
        context = get_log_context()
        assert "agent_id" not in context
        assert context["actor_id"] == "user-456"
        clear_log_context()

    def test_clear_log_context(self):
        update_log_context(agent_id="agent-123", actor_id="user-456")
        clear_log_context()
        context = get_log_context()
        assert context == {}

    def test_get_log_context_all(self):
        clear_log_context()
        update_log_context(agent_id="agent-123", actor_id="user-456")
        context = get_log_context()
        assert isinstance(context, dict)
        assert len(context) == 2
        clear_log_context()


class TestLogContextFilter:
    def test_filter_adds_context_to_record(self):
        clear_log_context()
        update_log_context(agent_id="agent-123", actor_id="user-456")

        log_filter = LogContextFilter()
        record = logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test.py",
            lineno=1,
            msg="Test message",
            args=(),
            exc_info=None,
        )

        result = log_filter.filter(record)
        assert result is True
        assert hasattr(record, "agent_id")
        assert record.agent_id == "agent-123"
        assert hasattr(record, "actor_id")
        assert record.actor_id == "user-456"
        clear_log_context()

    def test_filter_does_not_override_existing_attributes(self):
        clear_log_context()
        update_log_context(agent_id="agent-123")

        log_filter = LogContextFilter()
        record = logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test.py",
            lineno=1,
            msg="Test message",
            args=(),
            exc_info=None,
        )
        record.agent_id = "agent-999"

        log_filter.filter(record)
        assert record.agent_id == "agent-999"
        clear_log_context()


class TestLogContextIntegration:
    def test_json_formatter_includes_context(self):
        clear_log_context()
        update_log_context(agent_id="agent-123", actor_id="user-456")

        logger = logging.getLogger("test_logger")
        logger.setLevel(logging.INFO)

        handler = logging.StreamHandler(StringIO())
        handler.setFormatter(JSONFormatter())
        handler.addFilter(LogContextFilter())
        logger.addHandler(handler)

        log_stream = handler.stream
        logger.info("Test message")

        log_stream.seek(0)
        log_output = log_stream.read()

        log_data = json.loads(log_output)
        assert log_data["message"] == "Test message"
        assert log_data["agent_id"] == "agent-123"
        assert log_data["actor_id"] == "user-456"

        logger.removeHandler(handler)
        clear_log_context()

    def test_multiple_log_calls_with_changing_context(self):
        clear_log_context()
        logger = logging.getLogger("test_logger_2")
        logger.setLevel(logging.INFO)

        handler = logging.StreamHandler(StringIO())
        handler.setFormatter(JSONFormatter())
        handler.addFilter(LogContextFilter())
        logger.addHandler(handler)

        log_stream = handler.stream

        update_log_context(agent_id="agent-123")
        logger.info("First message")

        update_log_context(actor_id="user-456")
        logger.info("Second message")

        log_stream.seek(0)
        log_lines = log_stream.readlines()
        assert len(log_lines) == 2

        first_log = json.loads(log_lines[0])
        assert first_log["agent_id"] == "agent-123"
        assert "actor_id" not in first_log

        second_log = json.loads(log_lines[1])
        assert second_log["agent_id"] == "agent-123"
        assert second_log["actor_id"] == "user-456"

        logger.removeHandler(handler)
        clear_log_context()



from typing import AsyncGenerator

from letta.adapters.letta_llm_adapter import LettaLLMAdapter
from letta.errors import LLMError
from letta.helpers.datetime_helpers import get_utc_timestamp_ns
from letta.interfaces.anthropic_streaming_interface import AnthropicStreamingInterface
from letta.interfaces.openai_streaming_interface import OpenAIStreamingInterface
from letta.llm_api.llm_client_base import LLMClientBase
from letta.otel.tracing import log_attributes, safe_json_dumps, trace_method
from letta.schemas.enums import LLMCallType, ProviderType
from letta.schemas.letta_message import LettaMessage
from letta.schemas.llm_config import LLMConfig
from letta.schemas.provider_trace import BillingContext, ProviderTrace
from letta.schemas.user import User
from letta.settings import settings
from letta.utils import safe_create_task


class LettaLLMStreamAdapter(LettaLLMAdapter):
    """
    Adapter for handling streaming LLM requests with immediate token yielding.

    This adapter supports real-time streaming of tokens from the LLM, providing
    minimal time-to-first-token (TTFT) latency. It uses specialized streaming
    interfaces for different providers (OpenAI, Anthropic) to handle their
    specific streaming formats.
    """

    def __init__(
        self,
        llm_client: LLMClientBase,
        llm_config: LLMConfig,
        call_type: LLMCallType,
        agent_id: str | None = None,
        agent_tags: list[str] | None = None,
        run_id: str | None = None,
        org_id: str | None = None,
        user_id: str | None = None,
        billing_context: "BillingContext | None" = None,
    ) -> None:
        super().__init__(
            llm_client,
            llm_config,
            call_type=call_type,
            agent_id=agent_id,
            agent_tags=agent_tags,
            run_id=run_id,
            org_id=org_id,
            user_id=user_id,
            billing_context=billing_context,
        )
        self.interface: OpenAIStreamingInterface | AnthropicStreamingInterface | None = None

    async def invoke_llm(
        self,
        request_data: dict,
        messages: list,
        tools: list,
        use_assistant_message: bool,
        requires_approval_tools: list[str] = [],
        step_id: str | None = None,
        actor: User | None = None,
    ) -> AsyncGenerator[LettaMessage, None]:
        """
        Execute a streaming LLM request and yield tokens/chunks as they arrive.

        This adapter:
        1. Makes a streaming request to the LLM
        2. Yields chunks immediately for minimal TTFT
        3. Accumulates response data through the streaming interface
        4. Updates all instance variables after streaming completes
        """
        # Store request data
        self.request_data = request_data

        # Instantiate streaming interface
        if self.llm_config.model_endpoint_type in [ProviderType.anthropic, ProviderType.bedrock, ProviderType.minimax]:
            self.interface = AnthropicStreamingInterface(
                use_assistant_message=use_assistant_message,
                put_inner_thoughts_in_kwarg=self.llm_config.put_inner_thoughts_in_kwargs,
                requires_approval_tools=requires_approval_tools,
                run_id=self.run_id,
                step_id=step_id,
            )
        elif self.llm_config.model_endpoint_type in [ProviderType.openai, ProviderType.openrouter]:
            # For non-v1 agents, always use Chat Completions streaming interface
            self.interface = OpenAIStreamingInterface(
                use_assistant_message=use_assistant_message,
                is_openai_proxy=self.llm_config.provider_name == "lmstudio_openai",
                put_inner_thoughts_in_kwarg=self.llm_config.put_inner_thoughts_in_kwargs,
                messages=messages,
                tools=tools,
                requires_approval_tools=requires_approval_tools,
                run_id=self.run_id,
                step_id=step_id,
            )
        else:
            raise ValueError(f"Streaming not supported for provider {self.llm_config.model_endpoint_type}")

        # Extract optional parameters
        # ttft_span = kwargs.get('ttft_span', None)

        request_start_ns = get_utc_timestamp_ns()

        # Start the streaming request (map provider errors to common LLMError types)
        try:
            stream = await self.llm_client.stream_async(request_data, self.llm_config)
        except Exception as e:
            self.llm_request_finish_timestamp_ns = get_utc_timestamp_ns()
            latency_ms = int((self.llm_request_finish_timestamp_ns - request_start_ns) / 1_000_000)
            await self.llm_client.log_provider_trace_async(
                request_data=request_data,
                response_json=None,
                llm_config=self.llm_config,
                latency_ms=latency_ms,
                error_msg=str(e),
                error_type=type(e).__name__,
            )
            raise self.llm_client.handle_llm_error(e, llm_config=self.llm_config)

        # Process the stream and yield chunks immediately for TTFT
        # Wrap in error handling to convert provider errors to common LLMError types
        try:
            async for chunk in self.interface.process(stream):  # TODO: add ttft span
                # Yield each chunk immediately as it arrives
                yield chunk
        except Exception as e:
            self.llm_request_finish_timestamp_ns = get_utc_timestamp_ns()
            latency_ms = int((self.llm_request_finish_timestamp_ns - request_start_ns) / 1_000_000)
            await self.llm_client.log_provider_trace_async(
                request_data=request_data,
                response_json=None,
                llm_config=self.llm_config,
                latency_ms=latency_ms,
                error_msg=str(e),
                error_type=type(e).__name__,
            )
            if isinstance(e, LLMError):
                raise
            raise self.llm_client.handle_llm_error(e, llm_config=self.llm_config)

        # After streaming completes, extract the accumulated data
        self.llm_request_finish_timestamp_ns = get_utc_timestamp_ns()

        # Extract tool call from the interface
        try:
            self.tool_call = self.interface.get_tool_call_object()
        except ValueError:
            # No tool call, handle upstream
            self.tool_call = None

        # Extract reasoning content from the interface
        self.reasoning_content = self.interface.get_reasoning_content()

        # Extract usage statistics from the streaming interface
        self.usage = self.interface.get_usage_statistics()
        self.usage.step_count = 1

        # Store any additional data from the interface
        self.message_id = self.interface.letta_message_id

        # Log request and response data
        self.log_provider_trace(step_id=step_id, actor=actor)

    def supports_token_streaming(self) -> bool:
        return True

    @trace_method
    def log_provider_trace(self, step_id: str | None, actor: User | None) -> None:
        """
        Log provider trace data for telemetry purposes in a fire-and-forget manner.

        Creates an async task to log the request/response data without blocking
        the main execution flow. For streaming adapters, this includes the final
        tool call and reasoning content collected during streaming.

        Args:
            step_id: The step ID associated with this request for logging purposes
            actor: The user associated with this request for logging purposes
        """
        if step_id is None or actor is None:
            return

        response_json = {
            "content": {
                "tool_call": self.tool_call.model_dump_json() if self.tool_call else None,
                "reasoning": [content.model_dump_json() for content in self.reasoning_content],
            },
            "id": self.interface.message_id,
            "model": self.interface.model,
            "role": "assistant",
            # "stop_reason": "",
            # "stop_sequence": None,
            "type": "message",
            "usage": {
                "input_tokens": self.usage.prompt_tokens,
                "output_tokens": self.usage.completion_tokens,
            },
        }

        # Store response data for future reference
        self.response_data = response_json

        log_attributes(
            {
                "request_data": safe_json_dumps(self.request_data),
                "response_data": safe_json_dumps(response_json),
            }
        )

        if settings.track_provider_trace:
            safe_create_task(
                self.telemetry_manager.create_provider_trace_async(
                    actor=actor,
                    provider_trace=ProviderTrace(
                        request_json=self.request_data,
                        response_json=response_json,
                        step_id=step_id,
                        agent_id=self.agent_id,
                        agent_tags=self.agent_tags,
                        run_id=self.run_id,
                        call_type=self.call_type,
                        org_id=self.org_id,
                        user_id=self.user_id,
                        llm_config=self.llm_config.model_dump() if self.llm_config else None,
                    ),
                ),
                label="create_provider_trace",
            )



#!/usr/bin/env python3
"""
Data Analysis Module - Advanced statistical and machine learning operations
Contains various data processing and analysis functions for research purposes.
"""

import warnings
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Dict, Optional

import numpy as np
import pandas as pd


class AnalysisType(Enum):
    """Enumeration of different analysis types."""

    DESCRIPTIVE = "descriptive"
    CORRELATION = "correlation"
    REGRESSION = "regression"
    CLUSTERING = "clustering"
    TIME_SERIES = "time_series"


@dataclass
class AnalysisResult:
    """Container for analysis results."""

    analysis_type: AnalysisType
    timestamp: datetime
    metrics: Dict[str, float]
    metadata: Dict[str, any]
    success: bool = True
    error_message: Optional[str] = None


class DataPreprocessor:
    """
    Advanced data preprocessing utility class.
    Handles cleaning, transformation, and feature engineering.
    """

    def __init__(self, missing_threshold: float = 0.5):
        self.missing_threshold = missing_threshold
        self.transformations_applied = []

    def clean_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Comprehensive data cleaning pipeline.

        Args:
            df: Input DataFrame to clean

        Returns:
            Cleaned DataFrame
        """
        original_shape = df.shape

        # Remove columns with excessive missing values
        missing_ratios = df.isnull().sum() / len(df)
        cols_to_drop = missing_ratios[missing_ratios > self.missing_threshold].index
        df_cleaned = df.drop(columns=cols_to_drop)

        if len(cols_to_drop) > 0:
            self.transformations_applied.append(f"Dropped {len(cols_to_drop)} columns")

        # Handle remaining missing values
        numeric_cols = df_cleaned.select_dtypes(include=[np.number]).columns
        categorical_cols = df_cleaned.select_dtypes(include=["object"]).columns

        # Fill numeric missing values with median
        for col in numeric_cols:
            if df_cleaned[col].isnull().any():
                median_value = df_cleaned[col].median()
                df_cleaned[col].fillna(median_value, inplace=True)
                self.transformations_applied.append(f"Filled {col} with median")

        # Fill categorical missing values with mode
        for col in categorical_cols:
            if df_cleaned[col].isnull().any():
                mode_value = df_cleaned[col].mode().iloc[0] if not df_cleaned[col].mode().empty else "Unknown"
                df_cleaned[col].fillna(mode_value, inplace=True)
                self.transformations_applied.append(f"Filled {col} with mode")

        # Remove duplicates
        initial_rows = len(df_cleaned)
        df_cleaned = df_cleaned.drop_duplicates()
        duplicates_removed = initial_rows - len(df_cleaned)

        if duplicates_removed > 0:
            self.transformations_applied.append(f"Removed {duplicates_removed} duplicate rows")

        print(f"Data cleaning complete: {original_shape} -> {df_cleaned.shape}")
        return df_cleaned

    def engineer_features(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Create new features from existing data.

        Args:
            df: Input DataFrame

        Returns:
            DataFrame with engineered features
        """
        df_featured = df.copy()

        # Numeric feature engineering
        numeric_cols = df_featured.select_dtypes(include=[np.number]).columns

        if len(numeric_cols) >= 2:
            # Create interaction features
            for i, col1 in enumerate(numeric_cols):
                for col2 in numeric_cols[i + 1 :]:
                    df_featured[f"{col1}_{col2}_ratio"] = df_featured[col1] / (df_featured[col2] + 1e-8)
                    df_featured[f"{col1}_{col2}_sum"] = df_featured[col1] + df_featured[col2]

            self.transformations_applied.append("Created interaction features")

        # Binning continuous variables
        for col in numeric_cols:
            if df_featured[col].nunique() > 10:  # Only bin if many unique values
                df_featured[f"{col}_binned"] = pd.qcut(df_featured[col], q=5, labels=False, duplicates="drop")
                self.transformations_applied.append(f"Binned {col}")

        return df_featured


class StatisticalAnalyzer:
    """
    Statistical analysis and hypothesis testing utilities.
    """

    @staticmethod
    def descriptive_statistics(df: pd.DataFrame) -> AnalysisResult:
        """
        Calculate comprehensive descriptive statistics.

        Args:
            df: Input DataFrame

        Returns:
            AnalysisResult with descriptive metrics
        """
        try:
            numeric_df = df.select_dtypes(include=[np.number])

            if numeric_df.empty:
                return AnalysisResult(
                    analysis_type=AnalysisType.DESCRIPTIVE,
                    timestamp=datetime.now(),
                    metrics={},
                    metadata={},
                    success=False,
                    error_message="No numeric columns found",
                )

            metrics = {
                "mean_values": numeric_df.mean().to_dict(),
                "std_values": numeric_df.std().to_dict(),
                "median_values": numeric_df.median().to_dict(),
                "skewness": numeric_df.skew().to_dict(),
                "kurtosis": numeric_df.kurtosis().to_dict(),
                "correlation_with_target": None,  # Would need target column
            }

            metadata = {
                "total_rows": len(df),
                "total_columns": len(df.columns),
                "numeric_columns": len(numeric_df.columns),
                "missing_values": df.isnull().sum().to_dict(),
            }

            return AnalysisResult(analysis_type=AnalysisType.DESCRIPTIVE, timestamp=datetime.now(), metrics=metrics, metadata=metadata)

        except Exception as e:
            return AnalysisResult(
                analysis_type=AnalysisType.DESCRIPTIVE,
                timestamp=datetime.now(),
                metrics={},
                metadata={},
                success=False,
                error_message=str(e),
            )

    @staticmethod
    def correlation_analysis(df: pd.DataFrame, method: str = "pearson") -> AnalysisResult:
        """
        Perform correlation analysis between variables.

        Args:
            df: Input DataFrame
            method: Correlation method ('pearson', 'spearman', 'kendall')

        Returns:
            AnalysisResult with correlation metrics
        """
        try:
            numeric_df = df.select_dtypes(include=[np.number])

            if len(numeric_df.columns) < 2:
                return AnalysisResult(
                    analysis_type=AnalysisType.CORRELATION,
                    timestamp=datetime.now(),
                    metrics={},
                    metadata={},
                    success=False,
                    error_message="Need at least 2 numeric columns for correlation",
                )

            corr_matrix = numeric_df.corr(method=method)

            # Find highest correlations (excluding diagonal)
            corr_pairs = []
            for i in range(len(corr_matrix.columns)):
                for j in range(i + 1, len(corr_matrix.columns)):
                    col1, col2 = corr_matrix.columns[i], corr_matrix.columns[j]
                    corr_value = corr_matrix.iloc[i, j]
                    if not np.isnan(corr_value):
                        corr_pairs.append((col1, col2, abs(corr_value)))

            # Sort by correlation strength
            corr_pairs.sort(key=lambda x: x[2], reverse=True)

            metrics = {
                "correlation_matrix": corr_matrix.to_dict(),
                "highest_correlations": corr_pairs[:10],  # Top 10
                "method_used": method,
            }

            metadata = {"variables_analyzed": list(numeric_df.columns), "total_pairs": len(corr_pairs)}

            return AnalysisResult(analysis_type=AnalysisType.CORRELATION, timestamp=datetime.now(), metrics=metrics, metadata=metadata)

        except Exception as e:
            return AnalysisResult(
                analysis_type=AnalysisType.CORRELATION,
                timestamp=datetime.now(),
                metrics={},
                metadata={},
                success=False,
                error_message=str(e),
            )


class TimeSeriesAnalyzer:
    """
    Time series analysis and forecasting utilities.
    """

    def __init__(self, frequency: str = "D"):
        self.frequency = frequency
        self.models_fitted = {}

    def detect_seasonality(self, series: pd.Series) -> Dict[str, any]:
        """
        Detect seasonal patterns in time series data.

        Args:
            series: Time series data

        Returns:
            Dictionary with seasonality information
        """
        try:
            # Simple seasonality detection using autocorrelation
            autocorr_values = []
            for lag in range(1, min(len(series) // 2, 365)):
                if len(series) > lag:
                    autocorr = series.autocorr(lag=lag)
                    if not np.isnan(autocorr):
                        autocorr_values.append((lag, autocorr))

            # Find peaks in autocorrelation
            significant_lags = [(lag, corr) for lag, corr in autocorr_values if abs(corr) > 0.5]
            significant_lags.sort(key=lambda x: abs(x[1]), reverse=True)

            return {
                "seasonal_lags": significant_lags[:5],
                "strongest_seasonality": significant_lags[0] if significant_lags else None,
                "autocorrelation_values": autocorr_values,
            }

        except Exception as e:
            warnings.warn(f"Seasonality detection failed: {e}")
            return {"error": str(e)}

    def trend_analysis(self, series: pd.Series, window: int = 30) -> Dict[str, any]:
        """
        Analyze trend patterns in time series.

        Args:
            series: Time series data
            window: Rolling window size for trend calculation

        Returns:
            Dictionary with trend information
        """
        try:
            # Calculate rolling statistics
            rolling_mean = series.rolling(window=window).mean()
            rolling_std = series.rolling(window=window).std()

            # Simple trend detection
            first_third = rolling_mean.iloc[: len(rolling_mean) // 3].mean()
            last_third = rolling_mean.iloc[-len(rolling_mean) // 3 :].mean()

            trend_direction = "increasing" if last_third > first_third else "decreasing"
            trend_strength = abs(last_third - first_third) / first_third if first_third != 0 else 0

            return {
                "trend_direction": trend_direction,
                "trend_strength": trend_strength,
                "rolling_mean": rolling_mean.to_dict(),
                "rolling_std": rolling_std.to_dict(),
                "volatility": rolling_std.mean(),
            }

        except Exception as e:
            warnings.warn(f"Trend analysis failed: {e}")
            return {"error": str(e)}


def generate_sample_data(n_samples: int = 1000) -> pd.DataFrame:
    """
    Generate sample dataset for testing analysis functions.

    Args:
        n_samples: Number of samples to generate

    Returns:
        Sample DataFrame
    """
    np.random.seed(42)

    data = {
        "feature_1": np.random.normal(100, 15, n_samples),
        "feature_2": np.random.exponential(2, n_samples),
        "feature_3": np.random.uniform(0, 100, n_samples),
        "category": np.random.choice(["A", "B", "C"], n_samples),
        "timestamp": pd.date_range("2023-01-01", periods=n_samples, freq="D"),
    }

    # Add some correlation
    data["feature_4"] = data["feature_1"] * 0.7 + np.random.normal(0, 10, n_samples)

    # Add missing values
    missing_indices = np.random.choice(n_samples, size=int(0.05 * n_samples), replace=False)
    for idx in missing_indices:
        col = np.random.choice(["feature_1", "feature_2", "feature_3"])
        data[col][idx] = np.nan

    return pd.DataFrame(data)


def main():
    """
    Demonstration of the data analysis pipeline.
    """
    print("=== Data Analysis Pipeline Demo ===")

    # Generate sample data
    df = generate_sample_data(1000)
    print(f"Generated dataset with shape: {df.shape}")

    # Data preprocessing
    preprocessor = DataPreprocessor(missing_threshold=0.1)
    df_clean = preprocessor.clean_data(df)
    df_featured = preprocessor.engineer_features(df_clean)

    print(f"Applied transformations: {preprocessor.transformations_applied}")

    # Statistical analysis
    analyzer = StatisticalAnalyzer()

    # Descriptive statistics
    desc_result = analyzer.descriptive_statistics(df_featured)
    if desc_result.success:
        print(f"Descriptive analysis completed at {desc_result.timestamp}")
        print(f"Analyzed {desc_result.metadata['numeric_columns']} numeric columns")

    # Correlation analysis
    corr_result = analyzer.correlation_analysis(df_featured)
    if corr_result.success:
        print(f"Correlation analysis completed")
        print(f"Found {len(corr_result.metrics['highest_correlations'])} significant correlations")

    # Time series analysis
    ts_analyzer = TimeSeriesAnalyzer()
    time_series = df_clean.set_index("timestamp")["feature_1"]

    ts_analyzer.detect_seasonality(time_series)
    trend = ts_analyzer.trend_analysis(time_series)

    print(f"Time series trend: {trend.get('trend_direction', 'unknown')}")
    print(f"Volatility: {trend.get('volatility', 0):.2f}")


if __name__ == "__main__":
    main()



"""add instructions field to sources

Revision ID: 18e300709530
Revises: 878607e41ca4
Create Date: 2025-05-08 17:56:20.877183

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op
from letta.settings import settings

# revision identifiers, used by Alembic.
revision: str = "18e300709530"
down_revision: Union[str, None] = "878607e41ca4"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Skip this migration for SQLite
    if not settings.letta_pg_uri_no_default:
        return

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("sources", sa.Column("instructions", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # Skip this migration for SQLite
    if not settings.letta_pg_uri_no_default:
        return

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("sources", "instructions")
    # ### end Alembic commands ###



import logging
import os
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed

import matplotlib.pyplot as plt
import pandas as pd
import pytest
from dotenv import load_dotenv
from letta_client import Letta
from tqdm import tqdm

from letta.schemas.block import Block
from letta.schemas.embedding_config import EmbeddingConfig
from letta.schemas.llm_config import LLMConfig
from letta.services.block_manager import BlockManager

logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.WARNING)


# --- Server Management --- #


def _run_server():
    """Starts the Letta server in a background thread."""
    load_dotenv()
    from letta.server.rest_api.app import start_server

    start_server(debug=True)


@pytest.fixture(scope="session")
def server_url():
    """Ensures a server is running and returns its base URL."""
    url = os.getenv("LETTA_SERVER_URL", "http://localhost:8283")

    if not os.getenv("LETTA_SERVER_URL"):
        thread = threading.Thread(target=_run_server, daemon=True)
        thread.start()
        time.sleep(2)  # Allow server startup time

    return url


# --- Client Setup --- #


@pytest.fixture(scope="session")
def client(server_url):
    """Creates a REST client for testing."""
    client = Letta(base_url=server_url)
    yield client


@pytest.fixture()
def roll_dice_tool(client):
    def roll_dice():
        """
        Rolls a 6 sided die.

        Returns:
            str: The roll result.
        """
        return "Rolled a 10!"

    tool = client.tools.upsert_from_function(func=roll_dice)
    # Yield the created tool
    yield tool


@pytest.fixture()
def rethink_tool(client):
    def rethink_memory(agent_state: "AgentState", new_memory: str, target_block_label: str) -> str:  # type: ignore  # noqa: F821
        """
        Re-evaluate the memory in block_name, integrating new and updated facts.
        Replace outdated information with the most likely truths, avoiding redundancy with original memories.
        Ensure consistency with other memory blocks.

        Args:
            new_memory (str): The new memory with information integrated from the memory block. If there is no new information, then this should be the same as the content in the source block.
            target_block_label (str): The name of the block to write to.
        Returns:
            str: None is always returned as this function does not produce a response.
        """
        agent_state.memory.update_block_value(label=target_block_label, value=new_memory)
        return None

    tool = client.tools.upsert_from_function(func=rethink_memory)
    yield tool


@pytest.fixture
def default_block(default_user):
    """Fixture to create and return a default block."""
    block_manager = BlockManager()
    block_data = Block(
        label="default_label",
        value="Default Block Content",
        description="A default test block",
        limit=1000,
        metadata={"type": "test"},
    )
    block = block_manager.create_or_update_block(block_data, actor=default_user)
    yield block


@pytest.fixture(scope="function")
def agent_state(client, roll_dice_tool, weather_tool, rethink_tool):
    agent_state = client.agents.create(
        name=f"test_compl_{str(uuid.uuid4())[5:]}",
        tool_ids=[roll_dice_tool.id, weather_tool.id, rethink_tool.id],
        include_base_tools=True,
        memory_blocks=[
            {
                "label": "human",
                "value": "Name: Matt",
            },
            {
                "label": "persona",
                "value": "Friendly agent",
            },
        ],
        llm_config=LLMConfig.default_config(model_name="gpt-4o-mini"),
        embedding_config=EmbeddingConfig.default_config(provider="openai"),
    )
    yield agent_state
    client.agents.delete(agent_state.id)


# --- Load Test --- #


def create_agents_for_user(client, roll_dice_tool, rethink_tool, user_index: int) -> tuple:
    """Create agents and return E2E latencies in seconds along with user index."""
    # Setup blocks first
    num_blocks = 10
    blocks = []
    for i in range(num_blocks):
        block = client.blocks.create(
            label=f"user{user_index}_block{i}",
            value="Default Block Content",
            description="A default test block",
            limit=1000,
            metadata={"index": str(i)},
        )
        blocks.append(block)
    block_ids = [b.id for b in blocks]

    # Now create agents and track individual latencies
    agent_latencies = []
    num_agents_per_user = 100
    for i in range(num_agents_per_user):
        start_time = time.time()

        client.agents.create(
            name=f"user{user_index}_agent_{str(uuid.uuid4())[5:]}",
            tool_ids=[roll_dice_tool.id, rethink_tool.id],
            include_base_tools=True,
            memory_blocks=[
                {"label": "human", "value": "Name: Matt"},
                {"label": "persona", "value": "Friendly agent"},
            ],
            model="openai/gpt-4o",
            embedding_config=EmbeddingConfig.default_config(provider="openai"),
            block_ids=block_ids,
        )

        end_time = time.time()
        latency = end_time - start_time
        agent_latencies.append({"user_index": user_index, "agent_index": i, "latency": latency})

    return user_index, agent_latencies


def plot_agent_creation_latencies(latency_data):
    """
    Plot the distribution of agent creation latencies.

    Args:
        latency_data: List of dictionaries with latency information
    """
    # Convert to DataFrame for easier analysis
    df = pd.DataFrame(latency_data)

    # Overall latency distribution
    plt.figure(figsize=(12, 10))

    # Plot 1: Overall latency histogram
    plt.subplot(2, 2, 1)
    plt.hist(df["latency"], bins=30, alpha=0.7, color="blue")
    plt.title(f"Agent Creation Latency Distribution (n={len(df)})")
    plt.xlabel("Latency (seconds)")
    plt.ylabel("Frequency")
    plt.grid(True, alpha=0.3)

    # Plot 2: Latency by user (boxplot)
    plt.subplot(2, 2, 2)
    user_groups = df.groupby("user_index")
    plt.boxplot([group["latency"] for _, group in user_groups])
    plt.title("Latency Distribution by User")
    plt.xlabel("User Index")
    plt.ylabel("Latency (seconds)")
    plt.xticks(range(1, len(user_groups) + 1), sorted(df["user_index"].unique()))
    plt.grid(True, alpha=0.3)

    # Plot 3: Time series of latencies
    plt.subplot(2, 1, 2)
    for user_idx in sorted(df["user_index"].unique()):
        user_data = df[df["user_index"] == user_idx]
        plt.plot(user_data["agent_index"], user_data["latency"], marker=".", linestyle="-", alpha=0.7, label=f"User {user_idx}")

    plt.title("Agent Creation Latency Over Time")
    plt.xlabel("Agent Creation Sequence")
    plt.ylabel("Latency (seconds)")
    plt.legend(loc="upper right")
    plt.grid(True, alpha=0.3)

    # Add statistics as text
    stats_text = (
        f"Mean: {df['latency'].mean():.2f}s\n"
        f"Median: {df['latency'].median():.2f}s\n"
        f"Min: {df['latency'].min():.2f}s\n"
        f"Max: {df['latency'].max():.2f}s\n"
        f"Std Dev: {df['latency'].std():.2f}s"
    )
    plt.figtext(0.02, 0.02, stats_text, fontsize=10, bbox=dict(facecolor="white", alpha=0.8))

    plt.tight_layout()

    # Save the plot
    plot_file = f"agent_creation_latency_plot_{time.strftime('%Y%m%d_%H%M%S')}.png"
    plt.savefig(plot_file)
    plt.close()

    print(f"Latency plot saved to {plot_file}")

    # Return statistics for reporting
    return {
        "mean": df["latency"].mean(),
        "median": df["latency"].median(),
        "min": df["latency"].min(),
        "max": df["latency"].max(),
        "std": df["latency"].std(),
        "count": len(df),
        "plot_file": plot_file,
    }


@pytest.mark.slow
def test_parallel_create_many_agents(client, roll_dice_tool, rethink_tool):
    num_users = 7
    max_workers = min(num_users, 20)

    # To collect all latency data across users
    all_latency_data = []

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = {
            executor.submit(create_agents_for_user, client, roll_dice_tool, rethink_tool, user_idx): user_idx
            for user_idx in range(num_users)
        }

        with tqdm(total=num_users, desc="Creating agents") as pbar:
            for future in as_completed(futures):
                try:
                    user_idx, user_latencies = future.result()
                    all_latency_data.extend(user_latencies)

                    # Calculate and display per-user statistics
                    latencies = [data["latency"] for data in user_latencies]
                    avg_latency = sum(latencies) / len(latencies)
                    tqdm.write(f"[User {user_idx}] Completed {len(latencies)} agents")
                    tqdm.write(f"[User {user_idx}] Avg: {avg_latency:.2f}s, Min: {min(latencies):.2f}s, Max: {max(latencies):.2f}s")
                except Exception as e:
                    user_idx = futures[future]
                    tqdm.write(f"[User {user_idx}] Error during agent creation: {str(e)}")
                pbar.update(1)

    if all_latency_data:
        # Plot all collected latency data
        stats = plot_agent_creation_latencies(all_latency_data)

        print("\n===== Agent Creation Latency Statistics =====")
        print(f"Total agents created: {stats['count']}")
        print(f"Mean latency: {stats['mean']:.2f} seconds")
        print(f"Median latency: {stats['median']:.2f} seconds")
        print(f"Min latency: {stats['min']:.2f} seconds")
        print(f"Max latency: {stats['max']:.2f} seconds")
        print(f"Standard deviation: {stats['std']:.2f} seconds")
        print(f"Latency plot saved to: {stats['plot_file']}")
        print("============================================")



"""
Tests for ConversationManager.
"""

import pytest

from letta.orm.errors import NoResultFound
from letta.schemas.conversation import CreateConversation, UpdateConversation
from letta.server.server import SyncServer
from letta.services.conversation_manager import ConversationManager

# ======================================================================================================================
# ConversationManager Tests
# ======================================================================================================================


@pytest.fixture
def conversation_manager():
    """Create a ConversationManager instance."""
    return ConversationManager()


@pytest.mark.asyncio
async def test_create_conversation(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test creating a conversation."""
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test conversation"),
        actor=default_user,
    )

    assert conversation.id is not None
    assert conversation.agent_id == sarah_agent.id
    assert conversation.summary == "Test conversation"
    assert conversation.id.startswith("conv-")


@pytest.mark.asyncio
async def test_create_conversation_no_summary(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test creating a conversation without summary."""
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(),
        actor=default_user,
    )

    assert conversation.id is not None
    assert conversation.agent_id == sarah_agent.id
    assert conversation.summary is None


@pytest.mark.asyncio
async def test_get_conversation_by_id(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test retrieving a conversation by ID."""
    # Create a conversation
    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Retrieve it
    retrieved = await conversation_manager.get_conversation_by_id(
        conversation_id=created.id,
        actor=default_user,
    )

    assert retrieved.id == created.id
    assert retrieved.agent_id == created.agent_id
    assert retrieved.summary == created.summary


@pytest.mark.asyncio
async def test_get_conversation_not_found(conversation_manager, server: SyncServer, default_user):
    """Test retrieving a non-existent conversation raises error."""
    with pytest.raises(NoResultFound):
        await conversation_manager.get_conversation_by_id(
            conversation_id="conv-nonexistent",
            actor=default_user,
        )


@pytest.mark.asyncio
async def test_list_conversations(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test listing conversations for an agent."""
    # Create multiple conversations
    for i in range(3):
        await conversation_manager.create_conversation(
            agent_id=sarah_agent.id,
            conversation_create=CreateConversation(summary=f"Conversation {i}"),
            actor=default_user,
        )

    # List them
    conversations = await conversation_manager.list_conversations(
        agent_id=sarah_agent.id,
        actor=default_user,
    )

    assert len(conversations) == 3


@pytest.mark.asyncio
async def test_list_conversations_with_limit(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test listing conversations with a limit."""
    # Create multiple conversations
    for i in range(5):
        await conversation_manager.create_conversation(
            agent_id=sarah_agent.id,
            conversation_create=CreateConversation(summary=f"Conversation {i}"),
            actor=default_user,
        )

    # List with limit
    conversations = await conversation_manager.list_conversations(
        agent_id=sarah_agent.id,
        actor=default_user,
        limit=2,
    )

    assert len(conversations) == 2


@pytest.mark.asyncio
async def test_update_conversation(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test updating a conversation."""
    # Create a conversation
    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Original"),
        actor=default_user,
    )

    # Update it
    updated = await conversation_manager.update_conversation(
        conversation_id=created.id,
        conversation_update=UpdateConversation(summary="Updated summary"),
        actor=default_user,
    )

    assert updated.id == created.id
    assert updated.summary == "Updated summary"


@pytest.mark.asyncio
async def test_delete_conversation(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test soft deleting a conversation."""
    # Create a conversation
    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="To delete"),
        actor=default_user,
    )

    # Delete it
    await conversation_manager.delete_conversation(
        conversation_id=created.id,
        actor=default_user,
    )

    # Verify it's no longer accessible
    with pytest.raises(NoResultFound):
        await conversation_manager.get_conversation_by_id(
            conversation_id=created.id,
            actor=default_user,
        )


@pytest.mark.asyncio
async def test_delete_conversation_removes_from_list(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that soft-deleted conversations are excluded from list results."""
    # Create two conversations
    conv1 = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Keep me"),
        actor=default_user,
    )
    conv2 = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Delete me"),
        actor=default_user,
    )

    # Delete one
    await conversation_manager.delete_conversation(
        conversation_id=conv2.id,
        actor=default_user,
    )

    # List should only return the non-deleted conversation
    conversations = await conversation_manager.list_conversations(
        agent_id=sarah_agent.id,
        actor=default_user,
    )
    conv_ids = [c.id for c in conversations]
    assert conv1.id in conv_ids
    assert conv2.id not in conv_ids


@pytest.mark.asyncio
async def test_delete_conversation_double_delete_raises(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that deleting an already-deleted conversation raises NoResultFound."""
    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Delete me twice"),
        actor=default_user,
    )

    await conversation_manager.delete_conversation(
        conversation_id=created.id,
        actor=default_user,
    )

    # Second delete should raise
    with pytest.raises(NoResultFound):
        await conversation_manager.delete_conversation(
            conversation_id=created.id,
            actor=default_user,
        )


@pytest.mark.asyncio
async def test_update_deleted_conversation_raises(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that updating a soft-deleted conversation raises NoResultFound."""
    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Original"),
        actor=default_user,
    )

    await conversation_manager.delete_conversation(
        conversation_id=created.id,
        actor=default_user,
    )

    with pytest.raises(NoResultFound):
        await conversation_manager.update_conversation(
            conversation_id=created.id,
            conversation_update=UpdateConversation(summary="Should fail"),
            actor=default_user,
        )


@pytest.mark.asyncio
async def test_delete_conversation_excluded_from_summary_search(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that soft-deleted conversations are excluded from summary search results."""
    await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="alpha search term"),
        actor=default_user,
    )
    to_delete = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="alpha deleted term"),
        actor=default_user,
    )

    await conversation_manager.delete_conversation(
        conversation_id=to_delete.id,
        actor=default_user,
    )

    results = await conversation_manager.list_conversations(
        agent_id=sarah_agent.id,
        actor=default_user,
        summary_search="alpha",
    )
    result_ids = [c.id for c in results]
    assert to_delete.id not in result_ids
    assert len(results) == 1


@pytest.mark.asyncio
async def test_conversation_isolation_by_agent(conversation_manager, server: SyncServer, sarah_agent, charles_agent, default_user):
    """Test that conversations are isolated by agent."""
    # Create conversation for sarah_agent
    await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Sarah's conversation"),
        actor=default_user,
    )

    # Create conversation for charles_agent
    await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(summary="Charles's conversation"),
        actor=default_user,
    )

    # List for sarah_agent
    sarah_convos = await conversation_manager.list_conversations(
        agent_id=sarah_agent.id,
        actor=default_user,
    )
    assert len(sarah_convos) == 1
    assert sarah_convos[0].summary == "Sarah's conversation"

    # List for charles_agent
    charles_convos = await conversation_manager.list_conversations(
        agent_id=charles_agent.id,
        actor=default_user,
    )
    assert len(charles_convos) == 1
    assert charles_convos[0].summary == "Charles's conversation"


@pytest.mark.asyncio
async def test_conversation_isolation_by_organization(
    conversation_manager, server: SyncServer, sarah_agent, default_user, other_user_different_org
):
    """Test that conversations are isolated by organization."""
    # Create conversation
    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Other org user should not be able to access it
    with pytest.raises(NoResultFound):
        await conversation_manager.get_conversation_by_id(
            conversation_id=created.id,
            actor=other_user_different_org,
        )


# ======================================================================================================================
# Conversation Message Management Tests
# ======================================================================================================================


@pytest.mark.asyncio
async def test_add_messages_to_conversation(
    conversation_manager, server: SyncServer, sarah_agent, default_user, hello_world_message_fixture
):
    """Test adding messages to a conversation."""
    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Add the message to the conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[hello_world_message_fixture.id],
        actor=default_user,
    )

    # Verify message is in conversation
    message_ids = await conversation_manager.get_message_ids_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # create_conversation auto-creates a system message at position 0
    assert len(message_ids) == 2
    assert hello_world_message_fixture.id in message_ids


@pytest.mark.asyncio
async def test_get_messages_for_conversation(
    conversation_manager, server: SyncServer, sarah_agent, default_user, hello_world_message_fixture
):
    """Test getting full message objects from a conversation."""
    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Add the message
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[hello_world_message_fixture.id],
        actor=default_user,
    )

    # Get full messages
    messages = await conversation_manager.get_messages_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # create_conversation auto-creates a system message at position 0
    assert len(messages) == 2
    assert any(m.id == hello_world_message_fixture.id for m in messages)


@pytest.mark.asyncio
async def test_message_ordering_in_conversation(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that messages maintain their order in a conversation."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create multiple messages
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text=f"Message {i}")],
        )
        for i in range(3)
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add messages in order
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # Verify order is maintained
    retrieved_ids = await conversation_manager.get_message_ids_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # create_conversation auto-creates a system message at position 0,
    # so the user messages start at index 1
    assert len(retrieved_ids) == len(messages) + 1
    assert retrieved_ids[1:] == [m.id for m in messages]


@pytest.mark.asyncio
async def test_update_in_context_messages(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test updating which messages are in context."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text=f"Message {i}")],
        )
        for i in range(3)
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add all messages
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # Update to only keep first and last in context
    await conversation_manager.update_in_context_messages(
        conversation_id=conversation.id,
        in_context_message_ids=[messages[0].id, messages[2].id],
        actor=default_user,
    )

    # Verify only the selected messages are in context
    in_context_ids = await conversation_manager.get_message_ids_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    assert len(in_context_ids) == 2
    assert messages[0].id in in_context_ids
    assert messages[2].id in in_context_ids
    assert messages[1].id not in in_context_ids


@pytest.mark.asyncio
async def test_empty_conversation_message_ids(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test getting message IDs from a newly created conversation (has auto-created system message)."""
    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Empty"),
        actor=default_user,
    )

    # create_conversation auto-creates a system message at position 0,
    # so a newly created conversation has exactly one message
    message_ids = await conversation_manager.get_message_ids_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    assert len(message_ids) == 1


@pytest.mark.asyncio
async def test_list_conversation_messages(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test listing messages from a conversation as LettaMessages."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages with different roles
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="Hello!")],
        ),
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="assistant",
            content=[TextContent(text="Hi there!")],
        ),
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add messages to conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # List conversation messages (returns LettaMessages)
    letta_messages = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # create_conversation auto-creates a system message, so we get 3 total
    assert len(letta_messages) == 3
    # Check message types
    message_types = [m.message_type for m in letta_messages]
    assert "system_message" in message_types
    assert "user_message" in message_types
    assert "assistant_message" in message_types


@pytest.mark.asyncio
async def test_list_conversation_messages_pagination(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test pagination when listing conversation messages."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create multiple messages
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text=f"Message {i}")],
        )
        for i in range(5)
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add messages to conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # List with limit
    letta_messages = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        limit=2,
    )
    assert len(letta_messages) == 2

    # List with after cursor (get messages after the first one)
    letta_messages_after = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        after=messages[0].id,
    )
    assert len(letta_messages_after) == 4  # Should get messages 1-4


# ======================================================================================================================
# Isolated Blocks Tests
# ======================================================================================================================


@pytest.mark.asyncio
async def test_create_conversation_with_isolated_blocks(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test creating a conversation with isolated block labels."""
    # Get the agent's blocks to know what labels exist
    agent_state = await server.agent_manager.get_agent_by_id_async(charles_agent.id, default_user, include_relationships=["memory"])
    block_labels = [block.label for block in agent_state.memory.blocks]
    assert len(block_labels) > 0, "Agent should have at least one block"

    # Create conversation with isolated blocks
    first_label = block_labels[0]
    conversation = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(
            summary="Test with isolated blocks",
            isolated_block_labels=[first_label],
        ),
        actor=default_user,
    )

    assert conversation.id is not None
    assert conversation.agent_id == charles_agent.id
    assert len(conversation.isolated_block_ids) == 1

    # Verify the isolated block was created
    isolated_blocks = await conversation_manager.get_isolated_blocks_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )
    assert first_label in isolated_blocks
    assert isolated_blocks[first_label].label == first_label


@pytest.mark.asyncio
async def test_isolated_blocks_have_different_ids(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test that isolated blocks have different IDs from agent's original blocks."""
    # Get the agent's blocks
    agent_state = await server.agent_manager.get_agent_by_id_async(charles_agent.id, default_user, include_relationships=["memory"])
    original_block = agent_state.memory.blocks[0]

    # Create conversation with isolated block
    conversation = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(
            summary="Test isolated block IDs",
            isolated_block_labels=[original_block.label],
        ),
        actor=default_user,
    )

    # Get the isolated blocks
    isolated_blocks = await conversation_manager.get_isolated_blocks_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # Verify the isolated block has a different ID
    isolated_block = isolated_blocks[original_block.label]
    assert isolated_block.id != original_block.id
    assert isolated_block.label == original_block.label
    assert isolated_block.value == original_block.value  # Same initial value


@pytest.mark.asyncio
async def test_isolated_blocks_are_conversation_specific(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test that isolated blocks are specific to each conversation."""
    # Get the agent's first block label
    agent_state = await server.agent_manager.get_agent_by_id_async(charles_agent.id, default_user, include_relationships=["memory"])
    block_label = agent_state.memory.blocks[0].label

    # Create two conversations with the same isolated block label
    conv1 = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(
            summary="Conversation 1",
            isolated_block_labels=[block_label],
        ),
        actor=default_user,
    )

    conv2 = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(
            summary="Conversation 2",
            isolated_block_labels=[block_label],
        ),
        actor=default_user,
    )

    # Get isolated blocks for both conversations
    isolated_blocks_1 = await conversation_manager.get_isolated_blocks_for_conversation(
        conversation_id=conv1.id,
        actor=default_user,
    )
    isolated_blocks_2 = await conversation_manager.get_isolated_blocks_for_conversation(
        conversation_id=conv2.id,
        actor=default_user,
    )

    # Verify they have different block IDs
    block_1 = isolated_blocks_1[block_label]
    block_2 = isolated_blocks_2[block_label]
    assert block_1.id != block_2.id


@pytest.mark.asyncio
async def test_create_conversation_invalid_block_label(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test that creating a conversation with non-existent block label raises error."""
    from letta.errors import LettaInvalidArgumentError

    with pytest.raises(LettaInvalidArgumentError) as exc_info:
        await conversation_manager.create_conversation(
            agent_id=charles_agent.id,
            conversation_create=CreateConversation(
                summary="Test invalid label",
                isolated_block_labels=["nonexistent_block_label"],
            ),
            actor=default_user,
        )

    assert "nonexistent_block_label" in str(exc_info.value)


@pytest.mark.asyncio
async def test_apply_isolated_blocks_to_agent_state(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test that isolated blocks are correctly applied to agent state."""
    # Get the original agent state
    original_agent_state = await server.agent_manager.get_agent_by_id_async(
        charles_agent.id, default_user, include_relationships=["memory"]
    )
    original_block = original_agent_state.memory.blocks[0]

    # Create conversation with isolated block
    conversation = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(
            summary="Test apply isolated blocks",
            isolated_block_labels=[original_block.label],
        ),
        actor=default_user,
    )

    # Get fresh agent state
    agent_state = await server.agent_manager.get_agent_by_id_async(charles_agent.id, default_user, include_relationships=["memory"])

    # Apply isolated blocks
    modified_state = await conversation_manager.apply_isolated_blocks_to_agent_state(
        agent_state=agent_state,
        conversation_id=conversation.id,
        actor=default_user,
    )

    # Verify the block was replaced
    modified_block = modified_state.memory.get_block(original_block.label)
    assert modified_block.id != original_block.id
    assert modified_block.label == original_block.label
    assert modified_block.id in conversation.isolated_block_ids


@pytest.mark.asyncio
async def test_conversation_without_isolated_blocks(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that creating a conversation without isolated blocks works normally."""
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="No isolated blocks"),
        actor=default_user,
    )

    assert conversation.id is not None
    assert conversation.isolated_block_ids == []

    isolated_blocks = await conversation_manager.get_isolated_blocks_for_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )
    assert isolated_blocks == {}


@pytest.mark.asyncio
async def test_apply_no_isolated_blocks_preserves_state(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test that applying isolated blocks to a conversation without them preserves original state."""
    # Create conversation without isolated blocks
    conversation = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(summary="No isolated blocks"),
        actor=default_user,
    )

    # Get agent state
    agent_state = await server.agent_manager.get_agent_by_id_async(charles_agent.id, default_user, include_relationships=["memory"])
    original_block_ids = [block.id for block in agent_state.memory.blocks]

    # Apply isolated blocks (should be a no-op)
    modified_state = await conversation_manager.apply_isolated_blocks_to_agent_state(
        agent_state=agent_state,
        conversation_id=conversation.id,
        actor=default_user,
    )

    # Verify blocks are unchanged
    modified_block_ids = [block.id for block in modified_state.memory.blocks]
    assert original_block_ids == modified_block_ids


@pytest.mark.asyncio
async def test_delete_conversation_cleans_up_isolated_blocks(conversation_manager, server: SyncServer, charles_agent, default_user):
    """Test that deleting a conversation also hard-deletes its isolated blocks."""
    # Get the agent's first block label
    agent_state = await server.agent_manager.get_agent_by_id_async(charles_agent.id, default_user, include_relationships=["memory"])
    block_label = agent_state.memory.blocks[0].label

    # Create conversation with isolated block
    conversation = await conversation_manager.create_conversation(
        agent_id=charles_agent.id,
        conversation_create=CreateConversation(
            summary="Test delete cleanup",
            isolated_block_labels=[block_label],
        ),
        actor=default_user,
    )

    # Get the isolated block ID
    isolated_block_ids = conversation.isolated_block_ids
    assert len(isolated_block_ids) == 1
    isolated_block_id = isolated_block_ids[0]

    # Verify the isolated block exists
    isolated_block = await server.block_manager.get_block_by_id_async(isolated_block_id, default_user)
    assert isolated_block is not None

    # Delete the conversation
    await conversation_manager.delete_conversation(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # Verify the isolated block was hard-deleted
    deleted_block = await server.block_manager.get_block_by_id_async(isolated_block_id, default_user)
    assert deleted_block is None


# ======================================================================================================================
# list_conversation_messages with order/reverse Tests
# ======================================================================================================================


@pytest.mark.asyncio
async def test_list_conversation_messages_ascending_order(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test listing messages in ascending order (oldest first)."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages in a known order
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text=f"Message {i}")],
        )
        for i in range(3)
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add messages to conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # List messages in ascending order (reverse=False)
    letta_messages = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        reverse=False,
    )

    # create_conversation auto-creates a system message at position 0,
    # so we get 4 messages total (system + 3 user messages)
    assert len(letta_messages) == 4
    # First message is the auto-created system message; "Message 0" is second
    assert letta_messages[0].message_type == "system_message"
    assert "Message 0" in letta_messages[1].content


@pytest.mark.asyncio
async def test_list_conversation_messages_descending_order(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test listing messages in descending order (newest first)."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages in a known order
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text=f"Message {i}")],
        )
        for i in range(3)
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add messages to conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # List messages in descending order (reverse=True)
    letta_messages = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        reverse=True,
    )

    # create_conversation auto-creates a system message, so 4 total
    # First message should be "Message 2" (newest) in descending order
    assert len(letta_messages) == 4
    assert "Message 2" in letta_messages[0].content


@pytest.mark.asyncio
async def test_list_conversation_messages_with_group_id_filter(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test filtering messages by group_id."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages with different group_ids
    group_a_id = "group-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
    group_b_id = "group-bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"

    messages_group_a = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="Group A message 1")],
            group_id=group_a_id,
        ),
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="Group A message 2")],
            group_id=group_a_id,
        ),
    ]
    messages_group_b = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="Group B message 1")],
            group_id=group_b_id,
        ),
    ]

    created_a = await server.message_manager.create_many_messages_async(messages_group_a, actor=default_user)
    created_b = await server.message_manager.create_many_messages_async(messages_group_b, actor=default_user)

    # Add all messages to conversation
    all_message_ids = [m.id for m in created_a] + [m.id for m in created_b]
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=all_message_ids,
        actor=default_user,
    )

    # List messages filtered by group A
    messages_a = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        group_id=group_a_id,
    )

    assert len(messages_a) == 2
    for msg in messages_a:
        assert "Group A" in msg.content

    # List messages filtered by group B
    messages_b = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        group_id=group_b_id,
    )

    assert len(messages_b) == 1
    assert "Group B" in messages_b[0].content


@pytest.mark.asyncio
async def test_list_conversation_messages_no_group_id_returns_all(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that not providing group_id returns all messages."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages with different group_ids
    group_a_id = "group-aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
    group_b_id = "group-bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"

    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="Group A message")],
            group_id=group_a_id,
        ),
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="Group B message")],
            group_id=group_b_id,
        ),
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text="No group message")],
            group_id=None,
        ),
    ]
    messages = await server.message_manager.create_many_messages_async(pydantic_messages, actor=default_user)

    # Add all messages to conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # List all messages without group_id filter
    all_messages = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
    )

    # create_conversation auto-creates a system message, so 4 total
    assert len(all_messages) == 4


@pytest.mark.asyncio
async def test_list_conversation_messages_order_with_pagination(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that order affects pagination correctly."""
    from letta.schemas.letta_message_content import TextContent
    from letta.schemas.message import Message as PydanticMessage

    # Create a conversation
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test"),
        actor=default_user,
    )

    # Create messages
    pydantic_messages = [
        PydanticMessage(
            agent_id=sarah_agent.id,
            role="user",
            content=[TextContent(text=f"Message {i}")],
        )
        for i in range(5)
    ]
    messages = await server.message_manager.create_many_messages_async(
        pydantic_messages,
        actor=default_user,
    )

    # Add messages to conversation
    await conversation_manager.add_messages_to_conversation(
        conversation_id=conversation.id,
        agent_id=sarah_agent.id,
        message_ids=[m.id for m in messages],
        actor=default_user,
    )

    # Get first page in ascending order with limit
    page_asc = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        reverse=False,
        limit=2,
    )

    # Get first page in descending order with limit
    page_desc = await conversation_manager.list_conversation_messages(
        conversation_id=conversation.id,
        actor=default_user,
        reverse=True,
        limit=2,
    )

    # The first messages should be different
    assert page_asc[0].content != page_desc[0].content
    # In ascending, first is the auto-created system message, second is "Message 0"
    assert page_asc[0].message_type == "system_message"
    # In descending, first should be "Message 4"
    assert "Message 4" in page_desc[0].content


# ======================================================================================================================
# Model/Model Settings Override Tests
# ======================================================================================================================


@pytest.mark.asyncio
async def test_create_conversation_with_model(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test creating a conversation with a model override."""
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Test with model override", model="openai/gpt-4o"),
        actor=default_user,
    )

    assert conversation.id is not None
    assert conversation.model == "openai/gpt-4o"
    assert conversation.model_settings is None


@pytest.mark.asyncio
async def test_create_conversation_with_model_and_settings(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test creating a conversation with model and model_settings."""
    from letta.schemas.model import OpenAIModelSettings

    settings = OpenAIModelSettings(temperature=0.5)
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(
            summary="Test with settings",
            model="openai/gpt-4o",
            model_settings=settings,
        ),
        actor=default_user,
    )

    assert conversation.model == "openai/gpt-4o"
    assert conversation.model_settings is not None
    assert conversation.model_settings.temperature == 0.5


@pytest.mark.asyncio
async def test_create_conversation_without_model_override(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test creating a conversation without model override returns None for model fields."""
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="No override"),
        actor=default_user,
    )

    assert conversation.id is not None
    assert conversation.model is None
    assert conversation.model_settings is None


@pytest.mark.asyncio
async def test_update_conversation_set_model(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test updating a conversation to add a model override."""
    # Create without override
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="Original"),
        actor=default_user,
    )
    assert conversation.model is None

    # Update to add override
    updated = await conversation_manager.update_conversation(
        conversation_id=conversation.id,
        conversation_update=UpdateConversation(model="anthropic/claude-3-opus"),
        actor=default_user,
    )

    assert updated.model == "anthropic/claude-3-opus"


@pytest.mark.asyncio
async def test_update_conversation_preserves_model(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that updating summary preserves existing model override."""
    # Create with override
    conversation = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="With override", model="openai/gpt-4o"),
        actor=default_user,
    )
    assert conversation.model == "openai/gpt-4o"

    # Update summary only
    updated = await conversation_manager.update_conversation(
        conversation_id=conversation.id,
        conversation_update=UpdateConversation(summary="New summary"),
        actor=default_user,
    )

    assert updated.summary == "New summary"
    assert updated.model == "openai/gpt-4o"


@pytest.mark.asyncio
async def test_retrieve_conversation_includes_model(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that retrieving a conversation includes model/model_settings."""
    from letta.schemas.model import OpenAIModelSettings

    created = await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(
            summary="Retrieve test",
            model="openai/gpt-4o",
            model_settings=OpenAIModelSettings(temperature=0.7),
        ),
        actor=default_user,
    )

    retrieved = await conversation_manager.get_conversation_by_id(
        conversation_id=created.id,
        actor=default_user,
    )

    assert retrieved.model == "openai/gpt-4o"
    assert retrieved.model_settings is not None
    assert retrieved.model_settings.temperature == 0.7


@pytest.mark.asyncio
async def test_list_conversations_includes_model(conversation_manager, server: SyncServer, sarah_agent, default_user):
    """Test that listing conversations includes model fields."""
    await conversation_manager.create_conversation(
        agent_id=sarah_agent.id,
        conversation_create=CreateConversation(summary="List test", model="openai/gpt-4o"),
        actor=default_user,
    )

    conversations = await conversation_manager.list_conversations(
        agent_id=sarah_agent.id,
        actor=default_user,
    )

    assert len(conversations) >= 1
    conv_with_model = [c for c in conversations if c.summary == "List test"]
    assert len(conv_with_model) == 1
    assert conv_with_model[0].model == "openai/gpt-4o"


@pytest.mark.asyncio
async def test_create_conversation_schema_model_validation():
    """Test that CreateConversation validates model handle format."""
    from letta.errors import LettaInvalidArgumentError

    # Valid format should work
    create = CreateConversation(model="openai/gpt-4o")
    assert create.model == "openai/gpt-4o"

    # Invalid format should raise
    with pytest.raises(LettaInvalidArgumentError):
        CreateConversation(model="invalid-no-slash")


@pytest.mark.asyncio
async def test_update_conversation_schema_model_validation():
    """Test that UpdateConversation validates model handle format."""
    from letta.errors import LettaInvalidArgumentError

    # Valid format should work
    update = UpdateConversation(model="anthropic/claude-3-opus")
    assert update.model == "anthropic/claude-3-opus"

    # Invalid format should raise
    with pytest.raises(LettaInvalidArgumentError):
        UpdateConversation(model="no-slash")



"""add_run_completion_and_duration_to_agents_table

Revision ID: 9758adf8fdd3
Revises: 9556081ce65b
Create Date: 2025-06-18 18:22:31.135685

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op
from letta.settings import settings

# revision identifiers, used by Alembic.
revision: str = "9758adf8fdd3"
down_revision: Union[str, None] = "9556081ce65b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Skip this migration for SQLite
    if not settings.letta_pg_uri_no_default:
        return

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("agents", sa.Column("last_run_completion", sa.DateTime(timezone=True), nullable=True))
    op.add_column("agents", sa.Column("last_run_duration_ms", sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # Skip this migration for SQLite
    if not settings.letta_pg_uri_no_default:
        return

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("agents", "last_run_duration_ms")
    op.drop_column("agents", "last_run_completion")
    # ### end Alembic commands ###



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from letta.helpers.datetime_helpers import get_utc_time
from letta.schemas.enums import PrimitiveType
from letta.schemas.letta_base import OrmMetadataBase


class BillingContext(BaseModel):
    """Billing context for LLM request cost tracking."""

    plan_type: Optional[str] = Field(None, description="Subscription tier")
    cost_source: Optional[str] = Field(None, description="Cost source: 'quota' or 'credits'")
    customer_id: Optional[str] = Field(None, description="Customer ID for billing records")


class BaseProviderTrace(OrmMetadataBase):
    __id_prefix__ = PrimitiveType.PROVIDER_TRACE.value


class ProviderTrace(BaseProviderTrace):
    """
    Letta's internal representation of a provider trace.

    Attributes:
        id (str): The unique identifier of the provider trace.
        request_json (Dict[str, Any]): JSON content of the provider request.
        response_json (Dict[str, Any]): JSON content of the provider response.
        step_id (str): ID of the step that this trace is associated with.
        agent_id (str): ID of the agent that generated this trace.
        agent_tags (list[str]): Tags associated with the agent for filtering.
        call_type (str): Type of call (agent_step, summarization, etc.).
        run_id (str): ID of the run this trace is associated with.
        source (str): Source service that generated this trace (memgpt-server, lettuce-py).
        organization_id (str): The unique identifier of the organization.
        user_id (str): The unique identifier of the user who initiated the request.
        compaction_settings (Dict[str, Any]): Compaction/summarization settings (only for summarization calls).
        llm_config (Dict[str, Any]): LLM configuration used for this call (only for non-summarization calls).
        created_at (datetime): The timestamp when the object was created.
    """

    id: str = BaseProviderTrace.generate_id_field()
    request_json: Dict[str, Any] = Field(..., description="JSON content of the provider request")
    response_json: Dict[str, Any] = Field(..., description="JSON content of the provider response")
    step_id: Optional[str] = Field(None, description="ID of the step that this trace is associated with")

    # Telemetry context fields
    agent_id: Optional[str] = Field(None, description="ID of the agent that generated this trace")
    agent_tags: Optional[list[str]] = Field(None, description="Tags associated with the agent for filtering")
    call_type: Optional[str] = Field(None, description="Type of call (agent_step, summarization, etc.)")
    run_id: Optional[str] = Field(None, description="ID of the run this trace is associated with")
    source: Optional[str] = Field(None, description="Source service that generated this trace (memgpt-server, lettuce-py)")

    # v2 protocol fields
    org_id: Optional[str] = Field(None, description="ID of the organization")
    user_id: Optional[str] = Field(None, description="ID of the user who initiated the request")
    compaction_settings: Optional[Dict[str, Any]] = Field(None, description="Compaction/summarization settings (summarization calls only)")
    llm_config: Optional[Dict[str, Any]] = Field(None, description="LLM configuration used for this call (non-summarization calls only)")

    billing_context: Optional[BillingContext] = Field(None, description="Billing context from request headers")

    created_at: datetime = Field(default_factory=get_utc_time, description="The timestamp when the object was created.")


class ProviderTraceMetadata(BaseProviderTrace):
    """Metadata-only representation of a provider trace (no request/response JSON)."""

    id: str = BaseProviderTrace.generate_id_field()
    step_id: Optional[str] = Field(None, description="ID of the step that this trace is associated with")

    # Telemetry context fields
    agent_id: Optional[str] = Field(None, description="ID of the agent that generated this trace")
    agent_tags: Optional[list[str]] = Field(None, description="Tags associated with the agent for filtering")
    call_type: Optional[str] = Field(None, description="Type of call (agent_step, summarization, etc.)")
    run_id: Optional[str] = Field(None, description="ID of the run this trace is associated with")
    source: Optional[str] = Field(None, description="Source service that generated this trace (memgpt-server, lettuce-py)")

    # v2 protocol fields
    org_id: Optional[str] = Field(None, description="ID of the organization")
    user_id: Optional[str] = Field(None, description="ID of the user who initiated the request")

    created_at: datetime = Field(default_factory=get_utc_time, description="The timestamp when the object was created.")



from typing import List, Optional, Union

from mistralai import OCRPageObject

from letta.log import get_logger
from letta.otel.tracing import trace_method
from letta.services.file_processor.file_types import ChunkingStrategy, file_type_registry

logger = get_logger(__name__)


class LlamaIndexChunker:
    """LlamaIndex-based text chunking with automatic splitter selection"""

    # Conservative default chunk sizes for fallback scenarios
    DEFAULT_CONSERVATIVE_CHUNK_SIZE = 384
    DEFAULT_CONSERVATIVE_CHUNK_OVERLAP = 25

    def __init__(self, chunk_size: int = 512, chunk_overlap: int = 50, file_type: Optional[str] = None):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.file_type = file_type

        # Create appropriate parser based on file type
        self.parser = self._create_parser_for_file_type(file_type, chunk_size, chunk_overlap)

        # Log which parser was selected
        parser_name = type(self.parser).__name__
        logger.info(f"LlamaIndexChunker initialized with {parser_name} for file type: {file_type}")

    def _create_parser_for_file_type(self, file_type: Optional[str], chunk_size: int, chunk_overlap: int):
        """Create appropriate parser based on file type"""
        if not file_type:
            # Default fallback
            from llama_index.core.node_parser import SentenceSplitter

            return SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

        try:
            # Get chunking strategy from file type registry
            chunking_strategy = file_type_registry.get_chunking_strategy_by_mime_type(file_type)
            logger.debug(f"Chunking strategy for {file_type}: {chunking_strategy}")

            if chunking_strategy == ChunkingStrategy.CODE:
                from llama_index.core.node_parser import CodeSplitter

                return CodeSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

            elif chunking_strategy == ChunkingStrategy.DOCUMENTATION:
                if file_type in ["text/markdown", "text/x-markdown"]:
                    from llama_index.core.node_parser import MarkdownNodeParser

                    return MarkdownNodeParser()
                elif file_type in ["text/html"]:
                    from llama_index.core.node_parser import HTMLNodeParser

                    return HTMLNodeParser()
                else:
                    # Fall back to sentence splitter for other documentation
                    from llama_index.core.node_parser import SentenceSplitter

                    return SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

            elif chunking_strategy == ChunkingStrategy.STRUCTURED_DATA:
                if file_type in ["application/json", "application/jsonl"]:
                    from llama_index.core.node_parser import JSONNodeParser

                    return JSONNodeParser()
                else:
                    # Fall back to sentence splitter for other structured data
                    from llama_index.core.node_parser import SentenceSplitter

                    return SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

            else:
                # Default to sentence splitter for PROSE and LINE_BASED
                from llama_index.core.node_parser import SentenceSplitter

                return SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

        except Exception as e:
            logger.warning(f"Failed to create specialized parser for {file_type}: {str(e)}. Using default SentenceSplitter.")
            from llama_index.core.node_parser import SentenceSplitter

            return SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

    @trace_method
    def chunk_text(self, content: Union[OCRPageObject, str]) -> List[str]:
        """Chunk text using LlamaIndex splitter"""
        try:
            # Handle different input types
            if isinstance(content, OCRPageObject):
                # Extract markdown from OCR page object
                text_content = content.markdown
            else:
                # Assume it's a string
                text_content = content

            # Use the selected parser
            if hasattr(self.parser, "split_text"):
                # Most parsers have split_text method
                return self.parser.split_text(text_content)
            elif hasattr(self.parser, "get_nodes_from_documents"):
                # Some parsers need Document objects
                from llama_index.core import Document
                from llama_index.core.node_parser import SentenceSplitter

                document = Document(text=text_content)
                nodes = self.parser.get_nodes_from_documents([document])

                # Further split nodes that exceed chunk_size using SentenceSplitter
                final_chunks = []
                sentence_splitter = SentenceSplitter(chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap)

                for node in nodes:
                    if len(node.text) > self.chunk_size:
                        # Split oversized nodes with sentence splitter
                        sub_chunks = sentence_splitter.split_text(node.text)
                        final_chunks.extend(sub_chunks)
                    else:
                        final_chunks.append(node.text)

                return final_chunks
            else:
                # Fallback - try to call the parser directly
                return self.parser(text_content)

        except Exception as e:
            logger.error(f"Chunking failed with {type(self.parser).__name__}: {str(e)}")
            # Try fallback with SentenceSplitter
            try:
                logger.info("Attempting fallback to SentenceSplitter")
                from llama_index.core.node_parser import SentenceSplitter

                fallback_parser = SentenceSplitter(chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap)

                # Extract text content if needed
                if isinstance(content, OCRPageObject):
                    text_content = content.markdown
                else:
                    text_content = content

                return fallback_parser.split_text(text_content)
            except Exception as fallback_error:
                logger.error(f"Fallback chunking also failed: {str(fallback_error)}")
                raise e  # Raise the original error

    @trace_method
    def default_chunk_text(
        self, content: Union[OCRPageObject, str], chunk_size: int | None = None, chunk_overlap: int | None = None
    ) -> List[str]:
        """Chunk text using default SentenceSplitter regardless of file type with conservative defaults"""
        try:
            from llama_index.core.node_parser import SentenceSplitter

            # Use provided defaults or fallback to conservative values
            chunk_size = chunk_size if chunk_size is not None else self.DEFAULT_CONSERVATIVE_CHUNK_SIZE
            chunk_overlap = chunk_overlap if chunk_overlap is not None else self.DEFAULT_CONSERVATIVE_CHUNK_OVERLAP
            default_parser = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)

            # Handle different input types
            if isinstance(content, OCRPageObject):
                text_content = content.markdown
            else:
                text_content = content

            return default_parser.split_text(text_content)

        except Exception as e:
            logger.error(f"Default chunking failed: {str(e)}")
            raise



import os
import platform
import subprocess
import venv
from typing import TYPE_CHECKING, Dict, Optional

from datamodel_code_generator import DataModelType, PythonVersion
from datamodel_code_generator.model import get_data_model_types
from datamodel_code_generator.parser.jsonschema import JsonSchemaParser

from letta.log import get_logger
from letta.schemas.sandbox_config import LocalSandboxConfig

if TYPE_CHECKING:
    from letta.schemas.tool import Tool

logger = get_logger(__name__)


def find_python_executable(local_configs: LocalSandboxConfig) -> str:
    """
    Determines the Python executable path based on sandbox configuration and platform.
    Resolves any '~' (tilde) paths to absolute paths.

    Returns:
        str: Full path to the Python binary.
    """
    sandbox_dir = os.path.expanduser(local_configs.sandbox_dir)  # Expand tilde

    if not local_configs.use_venv:
        return "python.exe" if platform.system().lower().startswith("win") else "python3"

    venv_path = os.path.join(sandbox_dir, local_configs.venv_name)
    python_exec = (
        os.path.join(venv_path, "Scripts", "python.exe")
        if platform.system().startswith("Win")
        else os.path.join(venv_path, "bin", "python3")
    )

    if not os.path.isfile(python_exec):
        raise FileNotFoundError(f"Python executable not found: {python_exec}. Ensure the virtual environment exists.")

    return python_exec


def run_subprocess(command: list, env: Optional[Dict[str, str]] = None, fail_msg: str = "Command failed"):
    """
    Helper to execute a subprocess with logging and error handling.

    Args:
        command (list): The command to run as a list of arguments.
        env (dict, optional): The environment variables to use for the process.
        fail_msg (str): The error message to log in case of failure.

    Raises:
        RuntimeError: If the subprocess execution fails.
    """
    logger.info(f"Running command: {' '.join(command)}")
    try:
        result = subprocess.run(command, check=True, capture_output=True, text=True, env=env)
        logger.info(f"Command successful. Output:\n{result.stdout}")
        return result.stdout
    except subprocess.CalledProcessError as e:
        logger.error(f"{fail_msg}\nSTDOUT:\n{e.stdout}\nSTDERR:\n{e.stderr}")
        raise RuntimeError(f"{fail_msg}: {e.stderr.strip()}") from e
    except Exception as e:
        logger.error(f"{fail_msg}: {e}")
        raise RuntimeError(f"{fail_msg}: {e}")


def ensure_pip_is_up_to_date(python_exec: str, env: Optional[Dict[str, str]] = None):
    """
    Ensures pip, setuptools, and wheel are up to date before installing any other dependencies.

    Args:
        python_exec (str): Path to the Python executable to use.
        env (dict, optional): Environment variables to pass to subprocess.
    """
    run_subprocess(
        [python_exec, "-m", "pip", "install", "--upgrade", "pip", "setuptools", "wheel"],
        env=env,
        fail_msg="Failed to upgrade pip, setuptools, and wheel.",
    )


def install_pip_requirements_for_sandbox(
    local_configs: LocalSandboxConfig,
    upgrade: bool = True,
    user_install_if_no_venv: bool = False,
    env: Optional[Dict[str, str]] = None,
    tool: Optional["Tool"] = None,
):
    """
    Installs the specified pip requirements inside the correct environment (venv or system).
    Installs both sandbox-level and tool-specific pip requirements.
    """
    sandbox_dir = os.path.expanduser(local_configs.sandbox_dir)  # Expand tilde
    local_configs.sandbox_dir = sandbox_dir  # Update the object to store the absolute path

    python_exec = find_python_executable(local_configs)

    # If using a virtual environment, upgrade pip before installing dependencies.
    if local_configs.use_venv:
        ensure_pip_is_up_to_date(python_exec, env=env)

    # Collect all pip requirements
    all_packages = []

    # Add sandbox-level pip requirements
    if local_configs.pip_requirements:
        packages = [f"{req.name}=={req.version}" if req.version else req.name for req in local_configs.pip_requirements]
        all_packages.extend(packages)
        logger.debug(f"Added sandbox pip requirements: {packages}")

    # Add tool-specific pip requirements
    if tool and tool.pip_requirements:
        tool_packages = [str(req) for req in tool.pip_requirements]
        all_packages.extend(tool_packages)
        logger.debug(f"Added tool pip requirements for {tool.name}: {tool_packages}")

    if not all_packages:
        logger.debug("No pip requirements specified; skipping installation.")
        return

    # Construct pip install command
    pip_cmd = [python_exec, "-m", "pip", "install"]
    if upgrade:
        pip_cmd.append("--upgrade")
    pip_cmd += all_packages

    if user_install_if_no_venv and not local_configs.use_venv:
        pip_cmd.append("--user")

    # Enhanced error message for better debugging
    sandbox_packages = [f"{req.name}=={req.version}" if req.version else req.name for req in (local_configs.pip_requirements or [])]
    tool_packages = [str(req) for req in (tool.pip_requirements if tool and tool.pip_requirements else [])]

    error_details = []
    if sandbox_packages:
        error_details.append(f"sandbox requirements: {', '.join(sandbox_packages)}")
    if tool_packages:
        error_details.append(f"tool requirements: {', '.join(tool_packages)}")

    context = f" ({'; '.join(error_details)})" if error_details else ""
    fail_msg = f"Failed to install pip packages{context}. This may be due to package version incompatibility. Consider updating package versions or removing version constraints."

    run_subprocess(pip_cmd, env=env, fail_msg=fail_msg)


def create_venv_for_local_sandbox(sandbox_dir_path: str, venv_path: str, env: Dict[str, str], force_recreate: bool):
    """
    Creates a virtual environment for the sandbox. If force_recreate is True, deletes and recreates the venv.

    Args:
        sandbox_dir_path (str): Path to the sandbox directory.
        venv_path (str): Path to the virtual environment directory.
        env (dict): Environment variables to use.
        force_recreate (bool): If True, delete and recreate the virtual environment.
    """
    sandbox_dir_path = os.path.expanduser(sandbox_dir_path)
    venv_path = os.path.expanduser(venv_path)

    # If venv exists and force_recreate is True, delete it
    if force_recreate and os.path.isdir(venv_path):
        logger.warning(f"Force recreating virtual environment at: {venv_path}")
        import shutil

        shutil.rmtree(venv_path)

    # Create venv if it does not exist
    if not os.path.isdir(venv_path):
        logger.info(f"Creating new virtual environment at {venv_path}")
        venv.create(venv_path, with_pip=True)

    pip_path = os.path.join(venv_path, "bin", "pip")
    try:
        # Step 2: Upgrade pip
        logger.info("Upgrading pip in the virtual environment...")
        subprocess.run([pip_path, "install", "--upgrade", "pip"], env=env, check=True)

        # Step 3: Install packages from requirements.txt if available
        requirements_txt_path = os.path.join(sandbox_dir_path, "requirements.txt")
        if os.path.isfile(requirements_txt_path):
            logger.info(f"Installing packages from requirements file: {requirements_txt_path}")
            subprocess.run([pip_path, "install", "-r", requirements_txt_path], env=env, check=True)
            logger.info("Successfully installed packages from requirements.txt")
        else:
            logger.warning("No requirements.txt file found. Skipping package installation.")

    except subprocess.CalledProcessError as e:
        logger.error(f"Error while setting up the virtual environment: {e}")
        raise RuntimeError(f"Failed to set up the virtual environment: {e}")


def add_imports_and_pydantic_schemas_for_args(args_json_schema: dict) -> str:
    data_model_types = get_data_model_types(DataModelType.PydanticV2BaseModel, target_python_version=PythonVersion.PY_311)
    parser = JsonSchemaParser(
        str(args_json_schema),
        data_model_type=data_model_types.data_model,
        data_model_root_type=data_model_types.root_model,
        data_model_field_type=data_model_types.field_model,
        data_type_manager_type=data_model_types.data_type_manager,
        dump_resolve_reference_action=data_model_types.dump_resolve_reference_action,
    )
    result = parser.parse()
    return result


def prepare_local_sandbox(
    local_cfg: LocalSandboxConfig,
    env: Dict[str, str],
    force_recreate: bool = False,
) -> None:
    """
    Ensure the sandbox virtual-env is freshly created and that
    requirements are installed.  Uses your existing helpers.
    """
    sandbox_dir = os.path.expanduser(local_cfg.sandbox_dir)
    venv_path = os.path.join(sandbox_dir, local_cfg.venv_name)

    create_venv_for_local_sandbox(
        sandbox_dir_path=sandbox_dir,
        venv_path=venv_path,
        env=env,
        force_recreate=force_recreate,
    )

    install_pip_requirements_for_sandbox(
        local_cfg,
        upgrade=True,
        user_install_if_no_venv=False,
        env=env,
    )



from typing import List, Optional

from pydantic import Field

from letta.schemas.letta_base import LettaBase


class FileStats(LettaBase):
    """File statistics for metadata endpoint"""

    file_id: str = Field(..., description="Unique identifier of the file")
    file_name: str = Field(..., description="Name of the file")
    file_size: Optional[int] = Field(None, description="Size of the file in bytes")


class SourceStats(LettaBase):
    """Aggregated metadata for a source"""

    source_id: str = Field(..., description="Deprecated: Use `folder_id` field instead. Unique identifier of the source", deprecated=True)
    source_name: str = Field(..., description="Deprecated: Use `folder_name` field instead. Name of the source", deprecated=True)
    file_count: int = Field(0, description="Number of files in the source")
    total_size: int = Field(0, description="Total size of all files in bytes")
    files: List[FileStats] = Field(default_factory=list, description="List of file statistics")


class OrganizationSourcesStats(LettaBase):
    """Complete metadata response for organization sources"""

    total_sources: int = Field(0, description="Total number of sources")
    total_files: int = Field(0, description="Total number of files across all sources")
    total_size: int = Field(0, description="Total size of all files in bytes")
    sources: List[SourceStats] = Field(default_factory=list, description="List of source metadata")



from ...errors import LLMJSONParsingError
from ...helpers.json_helpers import json_dumps, json_loads
from ..json_parser import clean_json
from .wrapper_base import LLMChatCompletionWrapper


class ZephyrMistralWrapper(LLMChatCompletionWrapper):
    """
    Wrapper for Zephyr Alpha and Beta, Mistral 7B:
    https://huggingface.co/HuggingFaceH4/zephyr-7b-alpha
    https://huggingface.co/HuggingFaceH4/zephyr-7b-beta
    Note: this wrapper formats a prompt that only generates JSON, no inner thoughts
    """

    def __init__(
        self,
        simplify_json_content=True,
        clean_function_args=True,
        include_assistant_prefix=True,
        include_opening_brace_in_prefix=True,
        include_section_separators=False,
    ):
        self.simplify_json_content = simplify_json_content
        self.clean_func_args = clean_function_args
        self.include_assistant_prefix = include_assistant_prefix
        self.include_opening_brance_in_prefix = include_opening_brace_in_prefix
        self.include_section_separators = include_section_separators

    def chat_completion_to_prompt(self, messages, functions, function_documentation=None):
        """
        Zephyr prompt format:
            <|system|>
            </s>
            <|user|>
            {prompt}</s>
            <|assistant|>
        (source: https://huggingface.co/TheBloke/zephyr-7B-beta-GGUF#prompt-template-zephyr)
        """

        prompt = ""

        IM_END_TOKEN = "</s>"

        # System instructions go first
        assert messages[0]["role"] == "system"
        prompt += "<|system|>"
        prompt += f"\n{messages[0]['content']}"

        # Next is the functions preamble
        def create_function_description(schema):
            # airorobos style
            func_str = ""
            func_str += f"{schema['name']}:"
            func_str += f"\n  description: {schema['description']}"
            func_str += "\n  params:"
            for param_k, param_v in schema["parameters"]["properties"].items():
                # TODO we're ignoring type
                func_str += f"\n    {param_k}: {param_v['description']}"
            # TODO we're ignoring schema['parameters']['required']
            return func_str

        # prompt += f"\nPlease select the most suitable function and parameters from the list of available functions below, based on the user's input. Provide your response in JSON format."
        prompt += "\nPlease select the most suitable function and parameters from the list of available functions below, based on the ongoing conversation. Provide your response in JSON format."
        prompt += "\nAvailable functions:"
        if function_documentation is not None:
            prompt += f"\n{function_documentation}"
        else:
            for function_dict in functions:
                prompt += f"\n{create_function_description(function_dict)}"

        # Put functions INSIDE system message (TODO experiment with this)
        prompt += IM_END_TOKEN

        def create_function_call(function_call):
            airo_func_call = {
                "function": function_call["name"],
                "params": json_loads(function_call["arguments"]),
            }
            return json_dumps(airo_func_call, indent=2)

        for message in messages[1:]:
            assert message["role"] in ["user", "assistant", "function", "tool"], message

            if message["role"] == "user":
                if self.simplify_json_content:
                    try:
                        content_json = json_loads(message["content"])
                        content_simple = content_json["message"]
                        prompt += f"\n<|user|>\n{content_simple}{IM_END_TOKEN}"
                        # prompt += f"\nUSER: {content_simple}"
                    except Exception:
                        prompt += f"\n<|user|>\n{message['content']}{IM_END_TOKEN}"
                        # prompt += f"\nUSER: {message['content']}"
            elif message["role"] == "assistant":
                prompt += "\n<|assistant|>"
                if message["content"] is not None:
                    prompt += f"\n{message['content']}"
                # prompt += f"\nASSISTANT: {message['content']}"
                # need to add the function call if there was one
                if message.get("function_call"):
                    prompt += f"\n{create_function_call(message['function_call'])}"
                prompt += f"{IM_END_TOKEN}"
            elif message["role"] in ["function", "tool"]:
                # TODO find a good way to add this
                # prompt += f"\nASSISTANT: (function return) {message['content']}"
                prompt += "\n<|assistant|>"
                prompt += f"\nFUNCTION RETURN: {message['content']}"
                # prompt += f"\nFUNCTION RETURN: {message['content']}"
                continue
            else:
                raise ValueError(message)

        # Add a sep for the response
        # if self.include_section_separators:
        # prompt += "\n### RESPONSE"

        if self.include_assistant_prefix:
            # prompt += f"\nASSISTANT:"
            prompt += "\n<|assistant|>"
            if self.include_opening_brance_in_prefix:
                prompt += "\n{"

        return prompt

    def clean_function_args(self, function_name, function_args):
        """Some basic Letta-specific cleaning of function args"""
        cleaned_function_name = function_name
        cleaned_function_args = function_args.copy() if function_args is not None else {}

        if function_name == "send_message":
            # strip request_heartbeat
            cleaned_function_args.pop("request_heartbeat", None)

        # TODO more cleaning to fix errors LLM makes
        return cleaned_function_name, cleaned_function_args

    def output_to_chat_completion_response(self, raw_llm_output):
        """Turn raw LLM output into a ChatCompletion style response with:
        "message" = {
            "role": "assistant",
            "content": ...,
            "function_call": {
                "name": ...
                "arguments": {
                    "arg1": val1,
                    ...
                }
            }
        }
        """
        if self.include_opening_brance_in_prefix and raw_llm_output[0] != "{":
            raw_llm_output = "{" + raw_llm_output

        try:
            function_json_output = clean_json(raw_llm_output)
        except Exception as e:
            raise Exception(f"Failed to decode JSON from LLM output:\n{raw_llm_output} - error\n{str(e)}")
        try:
            function_name = function_json_output["function"]
            function_parameters = function_json_output["params"]
        except KeyError as e:
            raise LLMJSONParsingError(f"Received valid JSON from LLM, but JSON was missing fields: {str(e)}")

        if self.clean_func_args:
            function_name, function_parameters = self.clean_function_args(function_name, function_parameters)

        message = {
            "role": "assistant",
            "content": None,
            "function_call": {
                "name": function_name,
                "arguments": json_dumps(function_parameters),
            },
        }
        return message


class ZephyrMistralInnerMonologueWrapper(ZephyrMistralWrapper):
    """Still expect only JSON outputs from model, but add inner monologue as a field"""

    """
    Wrapper for Zephyr Alpha and Beta, Mistral 7B:
    https://huggingface.co/HuggingFaceH4/zephyr-7b-alpha
    https://huggingface.co/HuggingFaceH4/zephyr-7b-beta
    Note: this wrapper formats a prompt with inner thoughts included
    """

    def __init__(
        self,
        simplify_json_content=True,
        clean_function_args=True,
        include_assistant_prefix=True,
        include_opening_brace_in_prefix=True,
        include_section_separators=True,
    ):
        self.simplify_json_content = simplify_json_content
        self.clean_func_args = clean_function_args
        self.include_assistant_prefix = include_assistant_prefix
        self.include_opening_brance_in_prefix = include_opening_brace_in_prefix
        self.include_section_separators = include_section_separators

    def chat_completion_to_prompt(self, messages, functions, function_documentation=None):
        prompt = ""

        IM_END_TOKEN = "</s>"

        # System insturctions go first
        assert messages[0]["role"] == "system"
        prompt += messages[0]["content"]

        # Next is the functions preamble
        def create_function_description(schema, add_inner_thoughts=True):
            # airorobos style
            func_str = ""
            func_str += f"{schema['name']}:"
            func_str += f"\n  description: {schema['description']}"
            func_str += "\n  params:"
            if add_inner_thoughts:
                func_str += "\n    inner_thoughts: Deep inner monologue private to you only."
            for param_k, param_v in schema["parameters"]["properties"].items():
                # TODO we're ignoring type
                func_str += f"\n    {param_k}: {param_v['description']}"
            # TODO we're ignoring schema['parameters']['required']
            return func_str

        # prompt += f"\nPlease select the most suitable function and parameters from the list of available functions below, based on the user's input. Provide your response in JSON format."
        prompt += "\nPlease select the most suitable function and parameters from the list of available functions below, based on the ongoing conversation. Provide your response in JSON format."
        prompt += "\nAvailable functions:"
        if function_documentation is not None:
            prompt += f"\n{function_documentation}"
        else:
            for function_dict in functions:
                prompt += f"\n{create_function_description(function_dict)}"

        def create_function_call(function_call, inner_thoughts=None):
            airo_func_call = {
                "function": function_call["name"],
                "params": {
                    "inner_thoughts": inner_thoughts,
                    **json_loads(function_call["arguments"]),
                },
            }
            return json_dumps(airo_func_call, indent=2)

        # Add a sep for the conversation
        if self.include_section_separators:
            prompt += "\n<|user|>"

        # Last are the user/assistant messages
        for message in messages[1:]:
            assert message["role"] in ["user", "assistant", "function", "tool"], message

            if message["role"] == "user":
                if self.simplify_json_content:
                    try:
                        content_json = json_loads(message["content"])
                        content_simple = content_json["message"]
                        prompt += f"\n<|user|>\n{content_simple}{IM_END_TOKEN}"
                    except Exception:
                        prompt += f"\n<|user|>\n{message['content']}{IM_END_TOKEN}"
            elif message["role"] == "assistant":
                prompt += "\n<|assistant|>"
                # need to add the function call if there was one
                inner_thoughts = message["content"]
                if message.get("function_call"):
                    prompt += f"\n{create_function_call(message['function_call'], inner_thoughts=inner_thoughts)}"
            elif message["role"] in ["function", "tool"]:
                # TODO find a good way to add this
                # prompt += f"\nASSISTANT: (function return) {message['content']}"
                prompt += f"\nFUNCTION RETURN: {message['content']}"
                continue
            else:
                raise ValueError(message)

        # Add a sep for the response
        # if self.include_section_separators:
        #    prompt += "\n### RESPONSE"

        if self.include_assistant_prefix:
            prompt += "\n<|assistant|>"
            if self.include_opening_brance_in_prefix:
                prompt += "\n{"

        return prompt

    def clean_function_args(self, function_name, function_args):
        """Some basic Letta-specific cleaning of function args"""
        cleaned_function_name = function_name
        cleaned_function_args = function_args.copy() if function_args is not None else {}

        if function_name == "send_message":
            # strip request_heartbeat
            cleaned_function_args.pop("request_heartbeat", None)

        inner_thoughts = None
        if "inner_thoughts" in function_args:
            inner_thoughts = cleaned_function_args.pop("inner_thoughts")

        # TODO more cleaning to fix errors LLM makes
        return inner_thoughts, cleaned_function_name, cleaned_function_args

    def output_to_chat_completion_response(self, raw_llm_output):
        """Turn raw LLM output into a ChatCompletion style response with:
        "message" = {
            "role": "assistant",
            "content": ...,
            "function_call": {
                "name": ...
                "arguments": {
                    "arg1": val1,
                    ...
                }
            }
        }
        """
        if self.include_opening_brance_in_prefix and raw_llm_output[0] != "{":
            raw_llm_output = "{" + raw_llm_output

        try:
            function_json_output = clean_json(raw_llm_output)
        except Exception as e:
            raise Exception(f"Failed to decode JSON from LLM output:\n{raw_llm_output} - error\n{str(e)}")
        try:
            function_name = function_json_output["function"]
            function_parameters = function_json_output["params"]
        except KeyError as e:
            raise LLMJSONParsingError(f"Received valid JSON from LLM, but JSON was missing fields: {str(e)}")

        if self.clean_func_args:
            (
                inner_thoughts,
                function_name,
                function_parameters,
            ) = self.clean_function_args(function_name, function_parameters)

        message = {
            "role": "assistant",
            "content": inner_thoughts,
            "function_call": {
                "name": function_name,
                "arguments": json_dumps(function_parameters),
            },
        }
        return message



from typing import List, Optional

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.exc import NoResultFound

from letta.orm.agent import Agent as AgentModel
from letta.orm.block import Block as BlockModel
from letta.orm.errors import UniqueConstraintViolationError
from letta.orm.identity import Identity as IdentityModel
from letta.otel.tracing import trace_method
from letta.schemas.agent import AgentState
from letta.schemas.block import Block
from letta.schemas.enums import PrimitiveType
from letta.schemas.identity import (
    Identity as PydanticIdentity,
    IdentityCreate,
    IdentityProperty,
    IdentityType,
    IdentityUpdate,
    IdentityUpsert,
)
from letta.schemas.user import User as PydanticUser
from letta.server.db import db_registry
from letta.settings import DatabaseChoice, settings
from letta.utils import bounded_gather, decrypt_agent_secrets, enforce_types
from letta.validators import raise_on_invalid_id


class IdentityManager:
    @enforce_types
    @trace_method
    async def list_identities_async(
        self,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        identifier_key: Optional[str] = None,
        identity_type: Optional[IdentityType] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 50,
        ascending: bool = False,
        actor: PydanticUser = None,
    ) -> tuple[list[PydanticIdentity], Optional[str], bool]:
        """
        List identities with pagination metadata.

        Returns:
            Tuple of (identities, next_cursor, has_more)
        """
        async with db_registry.async_session() as session:
            filters = {"organization_id": actor.organization_id}
            if project_id:
                filters["project_id"] = project_id
            if identifier_key:
                filters["identifier_key"] = identifier_key
            if identity_type:
                filters["identity_type"] = identity_type

            # Request one more than limit to check if there are more pages
            query_limit = limit + 1 if limit else None

            identities = await IdentityModel.list_async(
                db_session=session,
                query_text=name,
                before=before,
                after=after,
                limit=query_limit,
                ascending=ascending,
                **filters,
            )

            # Check if we got more records than requested (meaning there are more pages)
            has_more = len(identities) > limit if limit else False
            if has_more:
                # Trim back to the requested limit
                identities = identities[:limit]

            # Get cursor for next page (ID of last item in current page)
            next_cursor = identities[-1].id if identities else None

            return [identity.to_pydantic() for identity in identities], next_cursor, has_more

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @trace_method
    async def get_identity_async(self, identity_id: str, actor: PydanticUser) -> PydanticIdentity:
        async with db_registry.async_session() as session:
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)
            return identity.to_pydantic()

    @enforce_types
    @trace_method
    async def create_identity_async(self, identity: IdentityCreate, actor: PydanticUser) -> PydanticIdentity:
        async with db_registry.async_session() as session:
            return await self._create_identity_async(db_session=session, identity=identity, actor=actor)

    async def _create_identity_async(self, db_session, identity: IdentityCreate, actor: PydanticUser) -> PydanticIdentity:
        new_identity = IdentityModel(**identity.model_dump(exclude={"agent_ids", "block_ids"}, exclude_unset=True))
        new_identity.organization_id = actor.organization_id

        # For SQLite compatibility: check for unique constraint violation manually
        # since SQLite doesn't support postgresql_nulls_not_distinct=True
        if settings.database_engine is DatabaseChoice.SQLITE:
            # Check if an identity with the same identifier_key, project_id, and organization_id exists
            query = select(IdentityModel).where(
                IdentityModel.identifier_key == new_identity.identifier_key,
                IdentityModel.project_id == new_identity.project_id,
                IdentityModel.organization_id == new_identity.organization_id,
            )
            result = await db_session.execute(query)
            existing_identity = result.scalar_one_or_none()
            if existing_identity is not None:
                raise UniqueConstraintViolationError(
                    f"A unique constraint was violated for Identity. "
                    f"An identity with identifier_key='{new_identity.identifier_key}', "
                    f"project_id='{new_identity.project_id}', and "
                    f"organization_id='{new_identity.organization_id}' already exists."
                )

        await self._process_relationship_async(
            db_session=db_session,
            identity=new_identity,
            relationship_name="agents",
            model_class=AgentModel,
            item_ids=identity.agent_ids,
            allow_partial=False,
        )
        await self._process_relationship_async(
            db_session=db_session,
            identity=new_identity,
            relationship_name="blocks",
            model_class=BlockModel,
            item_ids=identity.block_ids,
            allow_partial=False,
        )
        await new_identity.create_async(db_session=db_session, actor=actor)
        return new_identity.to_pydantic()

    @enforce_types
    @trace_method
    async def upsert_identity_async(self, identity: IdentityUpsert, actor: PydanticUser) -> PydanticIdentity:
        async with db_registry.async_session() as session:
            existing_identity = await IdentityModel.read_async(
                db_session=session,
                identifier_key=identity.identifier_key,
                project_id=identity.project_id,
                organization_id=actor.organization_id,
                actor=actor,
            )

            if existing_identity is None:
                return await self._create_identity_async(db_session=session, identity=IdentityCreate(**identity.model_dump()), actor=actor)
            else:
                identity_update = IdentityUpdate(
                    name=identity.name,
                    identifier_key=identity.identifier_key,
                    identity_type=identity.identity_type,
                    agent_ids=identity.agent_ids,
                    properties=identity.properties,
                )
                return await self._update_identity_async(
                    db_session=session, existing_identity=existing_identity, identity=identity_update, actor=actor, replace=True
                )

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @trace_method
    async def update_identity_async(
        self, identity_id: str, identity: IdentityUpdate, actor: PydanticUser, replace: bool = False
    ) -> PydanticIdentity:
        async with db_registry.async_session() as session:
            try:
                existing_identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)
            except NoResultFound:
                raise HTTPException(status_code=404, detail="Identity not found")
            if existing_identity.organization_id != actor.organization_id:
                raise HTTPException(status_code=403, detail="Forbidden")

            return await self._update_identity_async(
                db_session=session, existing_identity=existing_identity, identity=identity, actor=actor, replace=replace
            )

    async def _update_identity_async(
        self,
        db_session,
        existing_identity: IdentityModel,
        identity: IdentityUpdate,
        actor: PydanticUser,
        replace: bool = False,
    ) -> PydanticIdentity:
        if identity.identifier_key is not None:
            existing_identity.identifier_key = identity.identifier_key
        if identity.name is not None:
            existing_identity.name = identity.name
        if identity.identity_type is not None:
            existing_identity.identity_type = identity.identity_type
        if identity.properties is not None:
            if replace:
                existing_identity.properties = [prop.model_dump() for prop in identity.properties]
            else:
                new_properties = {old_prop["key"]: old_prop for old_prop in existing_identity.properties} | {
                    new_prop.key: new_prop.model_dump() for new_prop in identity.properties
                }
                existing_identity.properties = list(new_properties.values())

        if identity.agent_ids is not None:
            await self._process_relationship_async(
                db_session=db_session,
                identity=existing_identity,
                relationship_name="agents",
                model_class=AgentModel,
                item_ids=identity.agent_ids,
                allow_partial=False,
                replace=replace,
            )
        if identity.block_ids is not None:
            await self._process_relationship_async(
                db_session=db_session,
                identity=existing_identity,
                relationship_name="blocks",
                model_class=BlockModel,
                item_ids=identity.block_ids,
                allow_partial=False,
                replace=replace,
            )
        await existing_identity.update_async(db_session=db_session, actor=actor)
        return existing_identity.to_pydantic()

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @trace_method
    async def upsert_identity_properties_async(
        self, identity_id: str, properties: List[IdentityProperty], actor: PydanticUser
    ) -> PydanticIdentity:
        async with db_registry.async_session() as session:
            existing_identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)
            if existing_identity is None:
                raise HTTPException(status_code=404, detail="Identity not found")
            return await self._update_identity_async(
                db_session=session,
                existing_identity=existing_identity,
                identity=IdentityUpdate(properties=properties),
                actor=actor,
                replace=True,
            )

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @trace_method
    async def delete_identity_async(self, identity_id: str, actor: PydanticUser) -> None:
        async with db_registry.async_session() as session:
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)
            if identity is None:
                raise HTTPException(status_code=404, detail="Identity not found")
            if identity.organization_id != actor.organization_id:
                raise HTTPException(status_code=403, detail="Forbidden")
            await session.delete(identity)
            # context manager now handles commits
            # await session.commit()

    @enforce_types
    @trace_method
    async def size_async(
        self,
        actor: PydanticUser,
    ) -> int:
        """
        Get the total count of identities for the given user.
        """
        async with db_registry.async_session() as session:
            return await IdentityModel.size_async(db_session=session, actor=actor)

    async def _process_relationship_async(
        self,
        db_session,
        identity: PydanticIdentity,
        relationship_name: str,
        model_class,
        item_ids: List[str],
        allow_partial=False,
        replace=True,
    ):
        current_relationship = getattr(identity, relationship_name, [])
        if not item_ids:
            if replace:
                setattr(identity, relationship_name, [])
            return

        # Retrieve models for the provided IDs
        found_items = (await db_session.execute(select(model_class).where(model_class.id.in_(item_ids)))).scalars().all()

        # Validate all items are found if allow_partial is False
        if not allow_partial and len(found_items) != len(item_ids):
            missing = set(item_ids) - {item.id for item in found_items}
            raise NoResultFound(f"Items not found in agents: {missing}")

        if replace:
            # Replace the relationship
            setattr(identity, relationship_name, found_items)
        else:
            # Extend the relationship (only add new items)
            current_ids = {item.id for item in current_relationship}
            new_items = [item for item in found_items if item.id not in current_ids]
            current_relationship.extend(new_items)

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @trace_method
    async def list_agents_for_identity_async(
        self,
        identity_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 50,
        ascending: bool = False,
        include: List[str] = [],
        actor: PydanticUser = None,
    ) -> List[AgentState]:
        """
        Get all agents associated with the specified identity.
        """
        async with db_registry.async_session() as session:
            # First verify the identity exists and belongs to the user
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)
            if identity is None:
                raise HTTPException(status_code=404, detail=f"Identity with id={identity_id} not found")

            # Get agents associated with this identity with pagination
            agents = await AgentModel.list_async(
                db_session=session,
                before=before,
                after=after,
                limit=limit,
                ascending=ascending,
                identity_id=identity.id,
            )

            # Convert without decrypting to release DB connection before PBKDF2
            agents_encrypted = await bounded_gather(
                [agent.to_pydantic_async(include_relationships=[], include=include, decrypt=False) for agent in agents]
            )

        # Decrypt secrets outside session
        return await decrypt_agent_secrets(agents_encrypted)

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @trace_method
    async def list_blocks_for_identity_async(
        self,
        identity_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 50,
        ascending: bool = False,
        actor: PydanticUser = None,
    ) -> List[Block]:
        """
        Get all blocks associated with the specified identity.
        """
        async with db_registry.async_session() as session:
            # First verify the identity exists and belongs to the user
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)
            if identity is None:
                raise HTTPException(status_code=404, detail=f"Identity with id={identity_id} not found")

            # Get blocks associated with this identity with pagination
            blocks = await BlockModel.list_async(
                db_session=session,
                before=before,
                after=after,
                limit=limit,
                ascending=ascending,
                identity_id=identity.id,
            )
            return [block.to_pydantic() for block in blocks]

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @raise_on_invalid_id(param_name="agent_id", expected_prefix=PrimitiveType.AGENT)
    @trace_method
    async def attach_agent_async(self, identity_id: str, agent_id: str, actor: PydanticUser) -> None:
        """
        Attach an agent to an identity.
        """
        async with db_registry.async_session() as session:
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)

            agent = await AgentModel.read_async(db_session=session, identifier=agent_id, actor=actor)

            # Add agent to identity if not already attached
            if agent not in identity.agents:
                identity.agents.append(agent)
                await identity.update_async(db_session=session, actor=actor)

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @raise_on_invalid_id(param_name="agent_id", expected_prefix=PrimitiveType.AGENT)
    @trace_method
    async def detach_agent_async(self, identity_id: str, agent_id: str, actor: PydanticUser) -> None:
        """
        Detach an agent from an identity.
        """
        async with db_registry.async_session() as session:
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)

            agent = await AgentModel.read_async(db_session=session, identifier=agent_id, actor=actor)

            # Remove agent from identity if attached
            if agent in identity.agents:
                identity.agents.remove(agent)
                await identity.update_async(db_session=session, actor=actor)

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @raise_on_invalid_id(param_name="block_id", expected_prefix=PrimitiveType.BLOCK)
    @trace_method
    async def attach_block_async(self, identity_id: str, block_id: str, actor: PydanticUser) -> None:
        """
        Attach a block to an identity.
        """
        async with db_registry.async_session() as session:
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)

            block = await BlockModel.read_async(db_session=session, identifier=block_id, actor=actor)

            # Add block to identity if not already attached
            if block not in identity.blocks:
                identity.blocks.append(block)
                await identity.update_async(db_session=session, actor=actor)

    @enforce_types
    @raise_on_invalid_id(param_name="identity_id", expected_prefix=PrimitiveType.IDENTITY)
    @raise_on_invalid_id(param_name="block_id", expected_prefix=PrimitiveType.BLOCK)
    @trace_method
    async def detach_block_async(self, identity_id: str, block_id: str, actor: PydanticUser) -> None:
        """
        Detach a block from an identity.
        """
        async with db_registry.async_session() as session:
            identity = await IdentityModel.read_async(db_session=session, identifier=identity_id, actor=actor)

            block = await BlockModel.read_async(db_session=session, identifier=block_id, actor=actor)

            # Remove block from identity if attached
            if block in identity.blocks:
                identity.blocks.remove(block)
                await identity.update_async(db_session=session, actor=actor)



"""add stop reasons to steps and message error flag

Revision ID: cce9a6174366
Revises: 2c059cad97cc
Create Date: 2025-07-10 13:56:17.383612

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "cce9a6174366"
down_revision: Union[str, None] = "2c059cad97cc"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("messages", sa.Column("is_err", sa.Boolean(), nullable=True))

    # manually added to handle non-table creation enums
    stopreasontype = sa.Enum(
        "end_turn", "error", "invalid_tool_call", "max_steps", "no_tool_call", "tool_rule", "cancelled", name="stopreasontype"
    )
    stopreasontype.create(op.get_bind())
    op.add_column("steps", sa.Column("stop_reason", stopreasontype, nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("steps", "stop_reason")
    op.drop_column("messages", "is_err")

    stopreasontype = sa.Enum(name="stopreasontype")
    stopreasontype.drop(op.get_bind())
    # ### end Alembic commands ###



import json
import logging
import traceback
from datetime import datetime, timezone
from logging.config import dictConfig
from pathlib import Path
from sys import stdout
from typing import Any, Optional

from letta.log_context import get_log_context
from letta.settings import log_settings, settings, telemetry_settings

selected_log_level = logging.DEBUG if settings.debug else logging.INFO


class JSONFormatter(logging.Formatter):
    """
    Custom JSON formatter for structured logging with Datadog integration.

    Outputs logs in JSON format with fields compatible with Datadog log ingestion.
    Automatically includes trace correlation fields when Datadog tracing is enabled.

    Usage:
        Enable JSON logging by setting the environment variable:
            LETTA_LOGGING_JSON_LOGGING=true

        Add custom structured fields to logs using the 'extra' parameter:
            logger.info("User action", extra={"user_id": "123", "action": "login"})

        These fields will be automatically included in the JSON output and
        indexed by Datadog for filtering and analysis.

    Output format:
        {
            "timestamp": "2025-10-23T18:34:24.931739+00:00",
            "level": "INFO",
            "logger": "Letta.module",
            "message": "Log message",
            "module": "module_name",
            "function": "function_name",
            "line": 123,
            "dd.trace_id": "1234567890",  # Added when Datadog tracing is enabled
            "dd.span_id": "9876543210",   # Added when Datadog tracing is enabled
            "custom_field": "custom_value" # Any extra fields you provide
        }
    """

    def format(self, record: logging.LogRecord) -> str:
        """Format log record as JSON with Datadog-compatible fields."""
        # Base log structure
        log_data: dict[str, Any] = {
            "timestamp": datetime.fromtimestamp(record.created, tz=timezone.utc).isoformat(),
            "level": record.levelname,
            "logger": record.name,
            "message": record.getMessage(),
            "module": record.module,
            "function": record.funcName,
            "line": record.lineno,
        }

        # Add Datadog trace correlation if available
        # ddtrace automatically injects these attributes when logging is patched
        if hasattr(record, "dd.trace_id"):
            log_data["dd.trace_id"] = getattr(record, "dd.trace_id")
        if hasattr(record, "dd.span_id"):
            log_data["dd.span_id"] = getattr(record, "dd.span_id")
        if hasattr(record, "dd.service"):
            log_data["dd.service"] = getattr(record, "dd.service")
        if hasattr(record, "dd.env"):
            log_data["dd.env"] = getattr(record, "dd.env")
        if hasattr(record, "dd.version"):
            log_data["dd.version"] = getattr(record, "dd.version")

        # Add OpenTelemetry trace correlation (for OTEL → Datadog integration)
        try:
            from opentelemetry import trace

            span = trace.get_current_span()
            if span and span.get_span_context().is_valid:
                ctx = span.get_span_context()
                # Format trace_id and span_id as Datadog expects (decimal strings)
                log_data["dd.trace_id"] = str(ctx.trace_id)
                log_data["dd.span_id"] = str(ctx.span_id)
        except Exception:
            # Fail silently if OTEL is not available
            pass

        # Add exception info if present
        if record.exc_info:
            log_data["exception"] = {
                "type": record.exc_info[0].__name__ if record.exc_info[0] else None,
                "message": str(record.exc_info[1]) if record.exc_info[1] else None,
                "stacktrace": "".join(traceback.format_exception(*record.exc_info)),
            }

        # Add any extra fields from the log record
        # These are custom fields passed via logging.info("msg", extra={...})
        for key, value in record.__dict__.items():
            if key not in [
                "name",
                "msg",
                "args",
                "created",
                "filename",
                "funcName",
                "levelname",
                "levelno",
                "lineno",
                "module",
                "msecs",
                "message",
                "pathname",
                "process",
                "processName",
                "relativeCreated",
                "thread",
                "threadName",
                "exc_info",
                "exc_text",
                "stack_info",
                "dd_env",
                "dd_service",
            ] and not key.startswith("dd."):
                log_data[key] = value

        return json.dumps(log_data, default=str)


class DatadogEnvFilter(logging.Filter):
    """
    Logging filter that adds Datadog-specific attributes to log records.

    This enables log-trace correlation by injecting environment and service metadata
    that Datadog can use to link logs with traces and other telemetry data.
    """

    def filter(self, record: logging.LogRecord) -> bool:
        """Add Datadog attributes to log record if Datadog is enabled."""
        if telemetry_settings.enable_datadog:
            record.dd_env = settings.environment or "development"
            record.dd_service = telemetry_settings.datadog_service_name
        else:
            # Provide defaults to prevent attribute errors if filter is applied incorrectly
            record.dd_env = ""
            record.dd_service = ""
        return True


class LogContextFilter(logging.Filter):
    """
    Logging filter that enriches log records with request context.

    Injects context-specific attributes like actor_id, agent_id, org_id, etc.
    into log records. These attributes are stored in a context variable
    and automatically included in all log messages within that context.

    This enables correlation of logs with specific requests, agents, and users
    in monitoring systems like Datadog.

    Usage:
        from letta.log_context import set_log_context, update_log_context

        # Set a single context value
        set_log_context("agent_id", "agent-123")

        # Set multiple context values
        update_log_context(agent_id="agent-123", actor_id="user-456")

        # All subsequent logs will include these attributes
        logger.info("Processing request")
        # Output: {"message": "Processing request", "agent_id": "agent-123", "actor_id": "user-456", ...}
    """

    def filter(self, record: logging.LogRecord) -> bool:
        """Add request context attributes to log record."""
        context = get_log_context()
        for key, value in context.items():
            if not hasattr(record, key):
                setattr(record, key, value)
        return True


def _setup_logfile() -> "Path":
    """ensure the logger filepath is in place

    Returns: the logfile Path
    """
    logfile = Path(settings.letta_dir / "logs" / "Letta.log")
    logfile.parent.mkdir(parents=True, exist_ok=True)
    logfile.touch(exist_ok=True)
    return logfile


# Determine which formatter to use based on configuration
def _get_console_formatter() -> str:
    """Determine the appropriate console formatter based on settings."""
    if log_settings.json_logging:
        return "json"
    elif telemetry_settings.enable_datadog:
        return "datadog"
    else:
        return "no_datetime"


def _get_file_formatter() -> str:
    """Determine the appropriate file formatter based on settings."""
    if log_settings.json_logging:
        return "json"
    elif telemetry_settings.enable_datadog:
        return "datadog"
    else:
        return "standard"


# Logging configuration with optional Datadog integration and JSON support
DEVELOPMENT_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,  # Allow capturing from all loggers
    "formatters": {
        "standard": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"},
        "no_datetime": {"format": "%(name)s - %(levelname)s - %(message)s"},
        "datadog": {
            # Datadog-compatible format with key=value pairs for better parsing
            # ddtrace's log injection will add dd.trace_id, dd.span_id automatically when logging is patched
            "format": "%(asctime)s - %(name)s - %(levelname)s - [dd.env=%(dd_env)s dd.service=%(dd_service)s] - %(message)s"
        },
        "json": {
            # JSON formatter for structured logging with full Datadog integration
            "()": JSONFormatter,
        },
    },
    "filters": {
        "datadog_env": {
            "()": DatadogEnvFilter,
        },
        "log_context": {
            "()": LogContextFilter,
        },
    },
    "handlers": {
        "console": {
            "level": selected_log_level,
            "class": "logging.StreamHandler",
            "stream": stdout,
            "formatter": _get_console_formatter(),
            "filters": (["datadog_env"] if telemetry_settings.enable_datadog and not log_settings.json_logging else []) + ["log_context"],
        },
        "file": {
            "level": "DEBUG",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": _setup_logfile(),
            "maxBytes": 1024**2 * 10,  # 10 MB per file
            "backupCount": 3,  # Keep 3 backup files
            "formatter": _get_file_formatter(),
            "filters": (["datadog_env"] if telemetry_settings.enable_datadog and not log_settings.json_logging else []) + ["log_context"],
        },
    },
    "root": {  # Root logger handles all logs
        "level": logging.DEBUG if settings.debug else logging.INFO,
        "handlers": ["console", "file"],
    },
    "loggers": {
        "Letta": {
            "level": logging.DEBUG if settings.debug else logging.INFO,
            "propagate": True,  # Let logs bubble up to root
        },
        "uvicorn": {
            "level": "CRITICAL",
            "handlers": ["console"],
            "propagate": True,
        },
        # Reduce noise from ddtrace internal logging
        "ddtrace": {
            "level": "WARNING",
            "propagate": True,
        },
    },
}

# Configure logging once at module initialization to avoid performance overhead
dictConfig(DEVELOPMENT_LOGGING)


def get_logger(name: Optional[str] = None) -> "logging.Logger":
    """returns the project logger, scoped to a child name if provided
    Args:
        name: will define a child logger
    """
    parent_logger = logging.getLogger("Letta")
    if name:
        return parent_logger.getChild(name)
    return parent_logger
